# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging as python_logging
import signal
import sys

from oslo_config import cfg
from oslo_log import log as logging
import pbr.version

from neutron.common import config as n_config  # noqa

from networking_bagpipe.bagpipe_bgp.api import api
from networking_bagpipe.bagpipe_bgp.common import config
from networking_bagpipe.bagpipe_bgp.engine import exabgp_peer_worker
from networking_bagpipe.bagpipe_bgp.vpn import dataplane_drivers as drivers


LOG = logging.getLogger(__name__)


def setup_config():
    config.register()
    cfg.CONF(args=sys.argv[1:],
             project='bagpipe-bgp',
             default_config_files=['/etc/bagpipe-bgp/bgp.conf'],
             version=('%%(prog)s %s' %
                      pbr.version.VersionInfo('networking-bagpipe')
                      .release_string()))


BAGPIPE_BGP_MODULE = "networking_bagpipe.bagpipe_bgp"


def setup_logging():
    # even in debug mode we don't want to much talk from these
    extra_log_level_defaults = [
        '%s.engine.exabgp_peer_worker.exabgp=INFO' % BAGPIPE_BGP_MODULE,
        '%s.common.looking_glass=WARNING' % BAGPIPE_BGP_MODULE,
        '%s.engine.route_table_manager=INFO' % BAGPIPE_BGP_MODULE,
        'ovsdbapp.backend.ovs_idl.vlog=INFO',
    ]

    logging.set_defaults(default_log_levels=(logging.get_default_log_levels() +
                                             extra_log_level_defaults))

    logging.setup(cfg.CONF, "bagpipe-bgp")


def fix_log_file():
    # assist transition from past bagpipe-bgp version which were
    # using --log-file to specify the location of a file to configure logging
    if (cfg.CONF.log_file and cfg.CONF.log_file.endswith('.conf')):
        cfg.CONF.log_file = None
        return ("now using oslo.log, specifying a log configuration file "
                "should be done with --log-config-append")


def daemon_main():
    logging.register_options(cfg.CONF)

    setup_config()

    if cfg.CONF.action != "unset":
        LOG.warning("Running daemonized and using start/stop is not supported "
                    "anymore, use of systemd is your typical alternative")
        if cfg.CONF.action == "stop":
            sys.exit(-1)

    log_file_warn = fix_log_file()

    setup_logging()

    if log_file_warn:
        LOG.warning(log_file_warn)

    exabgp_peer_worker.setup_exabgp_env()

    try:
        LOG.info("Starting bagpipe-bgp...")
        pecan_api = api.PecanAPI()

        cfg.CONF.log_opt_values(LOG, logging.INFO)

        def stop(signum, _):
            LOG.info("Received signal %d, stopping...", signum)
            pecan_api.stop()
            LOG.info("All threads now stopped...")
            sys.exit(0)

        signal.signal(signal.SIGTERM, stop)
        signal.signal(signal.SIGINT, stop)

        pecan_api.run()
    except Exception:
        LOG.exception("Error while starting BGP daemon")


def cleanup_main():
    logging.register_options(cfg.CONF)

    setup_config()

    fix_log_file()

    setup_logging()

    python_logging.root.name = "[BgpDataplaneCleaner]"

    for vpn_type, dataplane_driver in (
            drivers.instantiate_dataplane_drivers().items()):
        LOG.info("Cleaning dataplane for %s...", vpn_type)
        dataplane_driver.reset_state()

    LOG.info("BGP component dataplanes have been cleaned up.")


if __name__ == '__main__':
    daemon_main()



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock

from networking_bagpipe.bagpipe_bgp import constants as consts

from networking_bagpipe.bagpipe_bgp.common import utils
from networking_bagpipe.bagpipe_bgp.vpn import manager
from networking_bagpipe.tests.unit.bagpipe_bgp import base as t


REDIRECTED_INSTANCE_ID1 = 'redirected-id1'
REDIRECTED_INSTANCE_ID2 = 'redirected-id2'

MAC = "00:00:de:ad:be:ef"
IP = "10.0.0.1/32"
BRIDGE_NAME = "br-test"
LOCAL_PORT = {'linuxif': 'tap1'}
VPN_EXT_ID = 1
GW_IP = "10.0.0.1"
GW_MASK = 24
VNID = 255


class MockVPNInstance(object):

    type = consts.EVPN

    def __init__(self, vpn_manager, dataplane_driver,
                 external_instance_id, instance_id, import_rts, export_rts,
                 gateway_ip, mask, readvertise, attract_traffic, fallback=None,
                 **kwargs):
        self.manager = vpn_manager

        self.external_instance_id = external_instance_id
        self.instance_type = self.__class__.__name__
        self.instance_id = instance_id

        self.import_rts = import_rts
        self.export_rts = export_rts

    def __repr__(self):
        return "%s:%s:%s" % (self.instance_type,
                             self.instance_id,
                             self.external_instance_id)

    @classmethod
    def validate_convert_attach_params(*args):
        pass

    @classmethod
    def validate_convert_detach_params(*args):
        pass

    def update_fallback(self, *args):
        pass

    def update_route_targets(self, *args):
        pass

    def vif_plugged(self, *args, **kwargs):
        pass

    def vif_unplugged(self, *args):
        pass

    def start(self):
        pass

    def stop_if_empty(self):
        pass

    def stop(self):
        pass

    def join(self):
        pass


class TestVPNManager(t.TestCase):

    def setUp(self):
        super(TestVPNManager, self).setUp()
        mock.patch("networking_bagpipe.bagpipe_bgp.vpn.dataplane_drivers."
                   "instantiate_dataplane_drivers",
                   return_value={
                       'evpn': mock.Mock(),
                       'ipvpn': mock.Mock()
                   }).start()
        self.manager = manager.VPNManager()

    def tearDown(self):
        super(TestVPNManager, self).tearDown()
        self.manager.stop()

    def test_redirect_traffic_single_instance(self):
        redirect_instance = self.manager.redirect_traffic_to_vpn(
            REDIRECTED_INSTANCE_ID1, consts.IPVPN, t._rt_to_string(t.RT5)
        )

        # Check some VPN manager and redirect instance lists consistency
        self.assertIn(
            manager.redirect_instance_extid(consts.IPVPN,
                                            t._rt_to_string(t.RT5)),
            self.manager.vpn_instances)
        self.assertIn(REDIRECTED_INSTANCE_ID1,
                      redirect_instance.redirected_instances)

    def test_redirect_traffic_multiple_instance(self):
        redirect_instance_1 = self.manager.redirect_traffic_to_vpn(
            REDIRECTED_INSTANCE_ID1, consts.IPVPN, t._rt_to_string(t.RT5)
        )
        redirect_instance_2 = self.manager.redirect_traffic_to_vpn(
            REDIRECTED_INSTANCE_ID2, consts.IPVPN, t._rt_to_string(t.RT5)
        )

        # Check that same redirect instance is returned
        self.assertEqual(redirect_instance_2, redirect_instance_1)
        # Check some VPN manager and redirect instance lists consistency
        self.assertIn(
            manager.redirect_instance_extid(consts.IPVPN,
                                            t._rt_to_string(t.RT5)),
            self.manager.vpn_instances)
        self.assertIn(REDIRECTED_INSTANCE_ID1,
                      redirect_instance_1.redirected_instances)
        self.assertIn(REDIRECTED_INSTANCE_ID2,
                      redirect_instance_1.redirected_instances)

    def test_stop_redirect_traffic_multiple_instance(self):
        redirect_instance = self.manager.redirect_traffic_to_vpn(
            REDIRECTED_INSTANCE_ID1, consts.IPVPN, t._rt_to_string(t.RT5)
        )
        self.manager.redirect_traffic_to_vpn(
            REDIRECTED_INSTANCE_ID2, consts.IPVPN, t._rt_to_string(t.RT5)
        )

        # Check some VPN manager and redirect instance lists consistency
        self.manager.stop_redirect_to_vpn(REDIRECTED_INSTANCE_ID2,
                                          consts.IPVPN, t._rt_to_string(t.RT5))

        self.assertNotIn(REDIRECTED_INSTANCE_ID2,
                         redirect_instance.redirected_instances)

        self.manager.stop_redirect_to_vpn(REDIRECTED_INSTANCE_ID1,
                                          consts.IPVPN, t._rt_to_string(t.RT5))

        self.assertTrue(not self.manager.vpn_instances)

    def test_plug_vif_to_vpn_with_forced_vni(self):
        with mock.patch.object(self.manager, "_get_vpn_instance",
                               return_value=(mock.Mock(), False)
                               ) as mock_get_vpn_instance, \
                mock.patch.object(utils, "convert_route_targets"):
            self.manager.plug_vif_to_vpn(vpn_instance_id=VPN_EXT_ID,
                                         vpn_type=consts.EVPN,
                                         import_rt=[t.RT1],
                                         export_rt=[t.RT1],
                                         mac_address=MAC,
                                         ip_address=IP,
                                         gateway_ip=GW_IP,
                                         local_port=LOCAL_PORT,
                                         linuxbr=BRIDGE_NAME,
                                         vni=VNID)
        mock_get_vpn_instance.assert_called_once_with(
            VPN_EXT_ID, consts.EVPN, mock.ANY, mock.ANY, GW_IP, mock.ANY,
            None, None, None, linuxbr=BRIDGE_NAME, vni=VNID)

    def test_plug_vif_to_vpn_without_forced_vni(self):
        with mock.patch.object(self.manager, "_get_vpn_instance",
                               return_value=(mock.Mock(), False)
                               ) as mock_get_vpn_instance, \
                mock.patch.object(utils, "convert_route_targets"):
            self.manager.plug_vif_to_vpn(vpn_instance_id=VPN_EXT_ID,
                                         vpn_type=consts.EVPN,
                                         import_rt=[t.RT1],
                                         export_rt=[t.RT1],
                                         mac_address=MAC,
                                         ip_address=IP,
                                         gateway_ip=GW_IP,
                                         local_port=LOCAL_PORT,
                                         linuxbr=BRIDGE_NAME)

        mock_get_vpn_instance.assert_called_once_with(
            VPN_EXT_ID, consts.EVPN, mock.ANY, mock.ANY, GW_IP, mock.ANY,
            None, None, None, linuxbr=BRIDGE_NAME)

    def test_get_vpn_instance_with_forced_vni(self):
        instannce, _ = self.manager._get_vpn_instance(VPN_EXT_ID,
                                                      consts.IPVPN,
                                                      [], [],
                                                      GW_IP, GW_MASK,
                                                      None, None,
                                                      vni=VNID)
        instannce.start()

        self.assertEqual(VNID, instannce.instance_label,
                         "VPN instance label should be forced to VNID")

    def test_get_vpn_instance_without_forced_vni(self):
        instannce, _ = self.manager._get_vpn_instance(VPN_EXT_ID,
                                                      consts.IPVPN,
                                                      [], [],
                                                      GW_IP, GW_MASK,
                                                      None, None)

        instannce.start()

        self.assertIsNot(0, instannce.instance_label,
                         "VPN instance label should be assigned locally")

    def test_instance_id_uniqueness(self):
        with mock.patch.object(manager.VPNManager, 'type2class',
                               {consts.IPVPN: MockVPNInstance,
                                consts.EVPN: MockVPNInstance
                                }):
            vpn_instance_unplug_args = dict(vpn_type=consts.EVPN,
                                            mac_address=MAC,
                                            ip_address=IP)
            vpn_instance_plug_args = dict(vpn_type=consts.EVPN,
                                          import_rts=['64512:74'],
                                          export_rts=[],
                                          mac_address=MAC,
                                          ip_address_prefix=IP,
                                          gateway_ip=GW_IP,
                                          local_port=LOCAL_PORT)
            BASE_VPN_EXT = "extid-"

            for i in (1, 2, 3, 4, 5):
                self.manager.plug_vif_to_vpn(
                    external_instance_id=BASE_VPN_EXT+str(i),
                    **vpn_instance_plug_args)

            for i in (2, 4):
                self.manager.unplug_vif_from_vpn(
                    external_instance_id=BASE_VPN_EXT+str(i),
                    **vpn_instance_unplug_args)

            for i in (6, 7, 8):
                self.manager.plug_vif_to_vpn(
                    external_instance_id=BASE_VPN_EXT+str(i),
                    **vpn_instance_plug_args)

            instance_ids = [i.instance_id
                            for i in self.manager.vpn_instances.values()]

            # ensure that each value is unique
            self.assertEqual(len(self.manager.vpn_instances.values()),
                             len(set(instance_ids)))

    def test_instance_id_max(self):
        with mock.patch.object(manager.VPNManager, 'type2class',
                               {consts.IPVPN: MockVPNInstance,
                                consts.EVPN: MockVPNInstance
                                }):
            self.manager.next_vpn_instance_id = 2**32 - 1

            self.manager.plug_vif_to_vpn(
                external_instance_id="dummy1",
                vpn_type=consts.EVPN,
                mac_address=MAC,
                import_rts=[],
                export_rts=[],
                ip_address_prefix=IP,
                local_port=LOCAL_PORT)

            self.assertRaises(
                manager.MaxInstanceIDReached,
                self.manager.plug_vif_to_vpn,
                external_instance_id="dummy2",
                vpn_type=consts.EVPN,
                import_rts=[],
                export_rts=[],
                mac_address=MAC,
                ip_address_prefix=IP,
                local_port=LOCAL_PORT)

    @mock.patch('networking_bagpipe.bagpipe_bgp.engine.bgp_manager.Manager')
    def test_manager_stop(self, mocked_bgp_manager):
        instance, _ = self.manager._get_vpn_instance(
            "TEST_VPN_INSTANCE", consts.IPVPN, [t.RT1], [t.RT1],
            "192.168.0.1", 24, {}, {})
        instance.start()

        self.manager.stop()
        self.assertTrue(not self.manager.vpn_instances)



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2017 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# NOTE(tmorin): unit tests for system interactions aren't that useful, which
# is why you will find very few things here. The code is is expected to be
# covered by tempest and fullstack test jobstores

from networking_bagpipe.bagpipe_bgp.vpn.ipvpn import mpls_ovs_dataplane
from networking_bagpipe.tests.unit.bagpipe_bgp import base as t


class TestMPLSOVSDataplaneDriver(t.TestCase):

    def _test_priority_for_prefix(self, prefix_list):
        # assuming that prefix_list is a list of prefixes of increasing length
        # check that each prefixes has a priority higher than the previous one
        previous_prio = 0
        previous_prefix = ""
        for prefix in prefix_list:
            prio = mpls_ovs_dataplane._priority_from_prefix(prefix)
            self.assertTrue(prio > previous_prio,
                            ("%s should have a priority higher than %s, "
                             "but 0x%x !> 0x%x") % (prefix, previous_prefix,
                                                    prio, previous_prio))
            previous_prio = prio
            previous_prefix = prefix

    def test_priority_for_prefix_v4(self):
        self._test_priority_for_prefix(
            ["0.0.0.0/0", "1.1.0.0/16", "2.2.2.0/24", "3.3.3.3/32"])

    def test_priority_for_prefix_v6(self):
        self._test_priority_for_prefix(
            ["::0/0", "2001:db8:8:4::2/64", "::1/128"])

    def test_fallback_priority(self):
        self.assertTrue(mpls_ovs_dataplane.FALLBACK_PRIORITY <
                        mpls_ovs_dataplane._priority_from_prefix("0.0.0.0/0"))



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2018 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading

from oslo_log import log as logging

from networking_bagpipe.bagpipe_bgp.common import looking_glass as lg
from networking_bagpipe.bagpipe_bgp.common import utils
from networking_bagpipe.bagpipe_bgp.engine import exa

from neutron_lib import exceptions

LOG = logging.getLogger(__name__)


class MaxIDReached(exceptions.NeutronException):
    message = "Could not allocate identifier, maximum (%(max)d) was reached"


class IDAllocator(lg.LookingGlassMixin):

    MIN = 0
    MAX = None

    def __init__(self):
        self.allocated_ids = dict()
        self.released_ids = set()
        self.current_id = self.MIN

        self.lock = threading.Lock()

    @utils.synchronized
    def get_new_id(self, description):
        if len(self.released_ids) > 0:
            new_id = self.released_ids.pop()
        else:
            if self.current_id > self.MAX:
                LOG.error("All the %d possible identifiers have been "
                          "allocated.", self.MAX)
                raise MaxIDReached(max=self.MAX)

            new_id = self.current_id
            self.current_id += 1
        self.allocated_ids[new_id] = description

        LOG.debug("Allocated identifier %d for '%s'", new_id, description)

        return new_id

    @utils.synchronized
    def release(self, id):
        if id in self.allocated_ids:
            LOG.debug("Released identifier %d ('%s')", id,
                      self.allocated_ids[id])
            del self.allocated_ids[id]
            self.released_ids.add(id)
        else:
            LOG.warn("Asked to release a non registered identifier: %d", id)

    def get_lg_local_info(self, prefix):
        return self.allocated_ids


class RDAllocator(IDAllocator):

    MAX = 2**16-1

    def __init__(self, prefix):
        super(RDAllocator, self).__init__()
        self.prefix = prefix

    def get_new_rd(self, description):
        new_id = self.get_new_id(description)

        return exa.RouteDistinguisher.fromElements(self.prefix, new_id)

    def release(self, rd):
        super(RDAllocator, self).release(int(str(rd).split(':')[1]))


class LabelAllocator(IDAllocator):

    MIN = 16
    MAX = 2**20-1

    def get_new_label(self, description):
        return self.get_new_id(description)



# Copyright (c) 2016 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
from distutils import spawn
import os
import random
import signal

import fixtures
from oslo_utils import fileutils

from neutron.agent.linux import async_process
from neutron.tests import base
from neutron.tests.fullstack import base as neutron_base
from neutron.tests.fullstack.resources import process as neutron_proc


class BagpipeBGPFixture(fixtures.Fixture):

    BAGPIPE_BGP = "bagpipe-bgp"

    def __init__(self, env_desc, host_desc, test_name,
                 bgp_cfg_fixture, namespace=None):
        super(BagpipeBGPFixture, self).__init__()
        self.env_desc = env_desc
        self.host_desc = host_desc
        self.test_name = test_name
        self.bgp_cfg_fixture = bgp_cfg_fixture
        self.namespace = namespace

    def _setUp(self):
        self.process_fixture = self.useFixture(neutron_proc.ProcessFixture(
            test_name=self.test_name,
            process_name=self.BAGPIPE_BGP,
            exec_name=self.BAGPIPE_BGP,
            config_filenames=[self.bgp_cfg_fixture.filename],
            namespace=self.namespace,
            kill_signal=signal.SIGTERM))


class BagpipeFakeRRProcessFixture(neutron_proc.ProcessFixture):

    def start(self):
        cmd = [spawn.find_executable(self.exec_name)]
        self.process = async_process.AsyncProcess(
            cmd, run_as_root=True, namespace=self.namespace
        )
        self.process.start()


class BagpipeFakeRRFixture(fixtures.Fixture):

    BAGPIPE_FAKERR = "bagpipe-fakerr"

    def __init__(self, env_desc, host_desc, test_name):
        super(BagpipeFakeRRFixture, self).__init__()
        self.env_desc = env_desc
        self.host_desc = host_desc
        self.test_name = test_name

    def _setUp(self):
        self.process_fixture = self.useFixture(BagpipeFakeRRProcessFixture(
            test_name=self.test_name,
            process_name=self.BAGPIPE_FAKERR,
            exec_name=self.BAGPIPE_FAKERR,
            config_filenames=None,
            kill_signal=signal.SIGTERM))


class GoBGPProcessFixture(neutron_proc.ProcessFixture):

    # NOTE(tmorin): stopping the process does not work yet for
    # GOBGPD_LOG = True, because get_root_helper_child_pid is not designed to
    # find the right child pid when things when an intermediate shell
    # is used
    # (using 'sh -c "exec gobgpd ..."' does not work either, gobgpd silently
    # stops right after startup for a reason I did not identify)
    GOBGPD_LOG = False

    def start(self):
        test_name = base.sanitize_log_path(self.test_name)

        log_dir = os.path.join(neutron_base.DEFAULT_LOG_DIR, test_name)
        fileutils.ensure_tree(log_dir, mode=0o755)

        timestamp = datetime.datetime.now().strftime("%Y-%m-%d--%H-%M-%S-%f")
        log_file = "%s/%s--%s.log" % (log_dir, self.process_name,
                                      timestamp)

        gobgpd_exec = spawn.find_executable(self.exec_name)
        if not gobgpd_exec:
            raise Exception("can't find gobgpd executable in PATH (%s, %s)" %
                            (self.exec_name,
                             os.environ['PATH']))

        cmd = [
            gobgpd_exec,
            '-t', 'json',
            '-f', self.config_filenames[0],
            '--log-level=debug',
            # we don't need this management API:
            '--api-hosts=0.0.0.0:%s' % random.randint(20000, 30000)
        ]

        if self.GOBGPD_LOG:
            cmd = ['sh', '-c', ('%s > %s 2>&1') % (' '.join(cmd), log_file)]

        self.process = async_process.AsyncProcess(
            cmd, namespace=self.namespace
        )
        self.process.start()


class GoBGPFixture(fixtures.Fixture):

    GOBGPD = "gobgpd"

    def __init__(self, env_desc, host_desc, test_name, gobgp_cfg_fixture):
        super(GoBGPFixture, self).__init__()
        self.env_desc = env_desc
        self.host_desc = host_desc
        self.test_name = test_name
        self.gobgp_cfg_fixture = gobgp_cfg_fixture

    def _setUp(self):
        config_filenames = [self.gobgp_cfg_fixture.filename]

        self.process_fixture = self.useFixture(GoBGPProcessFixture(
            test_name=self.test_name,
            process_name=self.GOBGPD,
            exec_name=self.GOBGPD,
            config_filenames=config_filenames,
            kill_signal=signal.SIGTERM))



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import inspect
import logging as python_logging
import select
import time

from exabgp.bgp import fsm as exa_fsm
from exabgp.bgp import message as exa_message
from exabgp.bgp.message import open as exa_open
from exabgp.bgp import neighbor as exa_neighbor
from exabgp import logger as exa_logger
from exabgp.protocol import family as exa_family
from exabgp import reactor as exa_reactor
from exabgp.reactor import peer as exa_peer
from oslo_config import cfg
from oslo_log import log as logging
import six

from networking_bagpipe.bagpipe_bgp.common import looking_glass as lg
from networking_bagpipe.bagpipe_bgp import engine
from networking_bagpipe.bagpipe_bgp.engine import bgp_peer_worker
from networking_bagpipe.bagpipe_bgp.engine import exa


LOG = logging.getLogger(__name__)


def setup_exabgp_env():
    # initialize/tweak ExaBGP config and log internals

    from exabgp.configuration.setup import environment
    environment.application = 'bagpipe-bgp'
    env = environment.setup(None)
    # tell exabgp to parse routes:
    env.log.routes = True

    # we "tweak" the internals of exabgp Logger, so that (a) it does not break
    # oslo_log and (b) it logs through oslo_log
    # decorating the original restart would be better...
    exa_logger.Logger._restart = exa_logger.Logger.restart

    def decorated_restart(f):
        @six.wraps(f)
        def restart_never_first(self, first):
            # we don't want exabgp to really ever do its first restart stuff
            # that resets the root logger handlers
            return f(self, False)
        return restart_never_first

    exa_logger.Logger.restart = decorated_restart(
        exa_logger.Logger.restart
        )

    exa_logger.Logger._syslog = logging.getLogger(__name__ + ".exabgp").logger

    # prevent exabgp Logger code from adding or removing handlers for
    # this logger
    def noop(handler):
        pass

    exa_logger.Logger._syslog.addHandler = noop
    exa_logger.Logger._syslog.removeHandler = noop

    # patch the exabgp logging method, to avoid formating all the
    # information twice
    # we have to use different signature for exabgp (change betweeen
    # 4.0.1 and 4.0.2)
    FORMAT_SIG_401 = ['timestamp', 'level', 'source', 'message']
    if inspect.getargspec(exa_logger.Logger._format).args == FORMAT_SIG_401:
        def patched_format(self, timestamp, level, source, message):
            if self.short:
                return message
            return "%-13s %s" % (source, message)
    else:  # 4.0.2 or higher
        def patched_format(self, message, source, level, timestamp=None):
            if self.short:
                return message
            return "%-13s %s" % (source, message)

    exa_logger.Logger._format = patched_format

    env.log.enable = True

    if LOG.logger.getEffectiveLevel():
        env.log.level = environment.syslog_value(
            python_logging.getLevelName(LOG.logger.getEffectiveLevel())
            )
    else:
        env.log.level = environment.syslog_value('INFO')
    env.log.all = True
    env.log.packets = True

    # monkey patch exabgp to work around exabgp issue #690
    # only relevant for exabgp 4.0.2
    # ( https://github.com/Exa-Networks/exabgp/issues/690 )
    if hasattr(exa_family.Family, 'size'):
        exa_family.Family.size[(exa_family.AFI.l2vpn,
                                exa_family.SAFI.evpn)] = ((4,), 0)


TRANSLATE_EXABGP_STATE = {exa_fsm.FSM.IDLE: bgp_peer_worker.FSM.Idle,
                          exa_fsm.FSM.ACTIVE: bgp_peer_worker.FSM.Active,
                          exa_fsm.FSM.CONNECT: bgp_peer_worker.FSM.Connect,
                          exa_fsm.FSM.OPENSENT: bgp_peer_worker.FSM.OpenSent,
                          exa_fsm.FSM.OPENCONFIRM:
                              bgp_peer_worker.FSM.OpenConfirm,
                          exa_fsm.FSM.ESTABLISHED:
                              bgp_peer_worker.FSM.Established,
                          }


class ExaBGPPeerWorker(bgp_peer_worker.BGPPeerWorker, lg.LookingGlassMixin):

    enabled_families = [(exa.AFI.ipv4, exa.SAFI.mpls_vpn),
                        # (exa.exa.AFI.ipv6, exa.SAFI.mpls_vpn),
                        (exa.AFI.l2vpn, exa.SAFI.evpn),
                        (exa.AFI.ipv4, exa.SAFI.flow_vpn)]

    def __init__(self, bgp_manager, peer_address):
        bgp_peer_worker.BGPPeerWorker.__init__(self, bgp_manager, peer_address)

        self.local_address = cfg.CONF.BGP.local_address
        self.peer_address = peer_address

        self.peer = None

        self.rtc_active = False
        self._active_families = []

    # hooks into BGPPeerWorker state changes

    def _stop_and_clean(self):
        super(ExaBGPPeerWorker, self)._stop_and_clean()

        self._active_families = []

        if self.peer is not None:
            self.log.info("Clearing peer")
            if self.peer.proto:
                self.peer.proto.close()
            self.peer.stop()
            self.peer = None

    def _to_established(self):
        super(ExaBGPPeerWorker, self)._to_established()

        if self.rtc_active:
            self.log.debug("RTC active, subscribing to all RTC routes")
            # subscribe to RTC routes, to be able to propagate them from
            # internal workers to this peer
            self._subscribe(exa.AFI.ipv4, exa.SAFI.rtc)
        else:
            self.log.debug("RTC inactive, subscribing to all active families")
            # if we don't use RTC with our peer, then we need to see events for
            # all routes of all active families, to be able to send them to him
            for (afi, safi) in self._active_families:
                self._subscribe(afi, safi)

    # implementation of BGPPeerWorker abstract methods

    def _initiate_connection(self):
        self.log.debug("Initiate ExaBGP connection to %s:%s from %s",
                       self.peer_address, cfg.CONF.BGP.bgp_port,
                       self.local_address)

        self.rtc_active = False

        neighbor = exa_neighbor.Neighbor()
        neighbor.make_rib()
        neighbor.router_id = exa_open.RouterID(self.local_address)
        neighbor.local_as = exa.ASN(cfg.CONF.BGP.my_as)
        # no support for eBGP yet:
        neighbor.peer_as = exa.ASN(cfg.CONF.BGP.my_as)
        neighbor.local_address = exa.IP.create(self.local_address)
        neighbor.md5_ip = exa.IP.create(self.local_address)
        neighbor.peer_address = exa.IP.create(self.peer_address)
        neighbor.hold_time = exa_open.HoldTime(
            bgp_peer_worker.DEFAULT_HOLDTIME)
        neighbor.connect = cfg.CONF.BGP.bgp_port
        neighbor.api = collections.defaultdict(list)
        neighbor.extended_message = False

        for afi_safi in self.enabled_families:
            neighbor.add_family(afi_safi)

        if cfg.CONF.BGP.enable_rtc:
            # how to test fot this ?
            neighbor.add_family((exa.AFI.ipv4, exa.SAFI.rtc))

        self.log.debug("Instantiate ExaBGP Peer")
        self.peer = exa_peer.Peer(neighbor, None)

        try:
            for action in self.peer._establish():
                self.fsm.state = TRANSLATE_EXABGP_STATE[
                    self.peer.fsm.state]

                if action == exa_peer.ACTION.LATER:
                    time.sleep(2)
                elif action == exa_peer.ACTION.NOW:
                    time.sleep(0.1)

                if self.should_stop:
                    self.log.debug("We're closing, raise StoppedException")
                    raise bgp_peer_worker.StoppedException()

                if action == exa_peer.ACTION.CLOSE:
                    self.log.debug("Socket status is CLOSE, "
                                   "raise InitiateConnectionException")
                    raise bgp_peer_worker.InitiateConnectionException(
                        "Socket is closed")
        except exa_peer.Interrupted:
            self.log.debug("Connect was interrupted, "
                           "raise InitiateConnectionException")
            raise bgp_peer_worker.InitiateConnectionException(
                "Connect was interrupted")
        except exa_message.Notify as e:
            self.log.debug("Notify: %s", e)
            if (e.code, e.subcode) == (1, 1):
                raise bgp_peer_worker.OpenWaitTimeout(str(e))
            else:
                raise Exception("Notify received: %s" % e)
        except exa_reactor.network.error.LostConnection as e:
            raise

        # check the capabilities of the session just established...

        self.protocol = self.peer.proto

        received_open = self.protocol.negotiated.received_open

        self._set_hold_time(self.protocol.negotiated.holdtime)

        mp_capabilities = received_open.capabilities.get(
            exa_open.capability.Capability.CODE.MULTIPROTOCOL, [])

        # check that our peer advertized at least mpls_vpn and evpn
        # capabilities
        self._active_families = []
        for (afi, safi) in (self.__class__.enabled_families +
                            [(exa.AFI.ipv4, exa.SAFI.rtc)]):
            if (afi, safi) not in mp_capabilities:
                if (((afi, safi) != (exa.AFI.ipv4, exa.SAFI.rtc)) or
                        cfg.CONF.BGP.enable_rtc):
                    self.log.warning("Peer does not advertise (%s,%s) "
                                     "capability", afi, safi)
            else:
                self.log.info(
                    "Family (%s,%s) successfully negotiated with peer %s",
                    afi, safi, self.peer_address)
                self._active_families.append((afi, safi))

        if len(self._active_families) == 0:
            self.log.error("No family was negotiated for VPN routes")

        self.rtc_active = False

        if cfg.CONF.BGP.enable_rtc:
            if (exa.AFI.ipv4, exa.SAFI.rtc) in mp_capabilities:
                self.log.info(
                    "RTC successfully enabled with peer %s", self.peer_address)
                self.rtc_active = True
            else:
                self.log.warning(
                    "enable_rtc True but peer not configured for RTC")

    def _receive_loop_fun(self):

        try:
            select.select([self.protocol.connection.io], [], [], 2)

            if not self.protocol.connection:
                raise Exception("lost connection")

            message = six.next(self.protocol.read_message())

            if message.ID != exa_message.NOP.ID:
                self.log.debug("protocol read message: %s", message)
        except exa_message.Notification as e:
            self.log.error("Notification: %s", e)
            return 2
        except exa_reactor.network.error.LostConnection as e:
            self.log.warning("Lost connection while waiting for message: %s",
                             e)
            return 2
        except TypeError as e:
            self.log.error("Error while reading BGP message: %s", e)
            return 2
        except Exception as e:
            self.log.error("Error while reading BGP message: %s", e)
            raise

        if message.ID == exa_message.NOP.ID:
            return 1
        if message.ID == exa_message.Update.ID:
            if self.fsm.state != bgp_peer_worker.FSM.Established:
                raise Exception("Update received but not in Established state")
            # more below
        elif message.ID == exa_message.KeepAlive.ID:
            self.enqueue(bgp_peer_worker.KEEP_ALIVE_RECEIVED)
            self.log.debug("Received message: %s", message)
        else:
            self.log.warning("Received unexpected message: %s", message)

        if isinstance(message, exa_message.Update):
            if message.nlris:
                for nlri in message.nlris:
                    if nlri.action == exa_message.IN.ANNOUNCED:
                        action = engine.RouteEvent.ADVERTISE
                    elif nlri.action == exa_message.IN.WITHDRAWN:
                        action = engine.RouteEvent.WITHDRAW
                    else:
                        raise Exception("should not be reached (action:%s)",
                                        nlri.action)
                    self._process_received_route(action, nlri,
                                                 message.attributes)
        return 1

    def _process_received_route(self, action, nlri, attributes):
        self.log.info("Received route: %s, %s", nlri, attributes)

        route_entry = engine.RouteEntry(nlri, None, attributes)

        if action == exa_message.IN.ANNOUNCED:
            self._advertise_route(route_entry)
        elif action == exa_message.IN.WITHDRAWN:
            self._withdraw_route(route_entry)
        else:
            raise Exception("unsupported action ??? (%s)" % action)

        # TODO(tmmorin): move RTC code out-of the peer-specific code
        if (nlri.afi, nlri.safi) == (exa.AFI.ipv4, exa.SAFI.rtc):
            self.log.info("Received an RTC route")

            if nlri.rt is None:
                self.log.info("Received RTC is a wildcard")

            # the semantic of RTC routes does not distinguish between AFI/SAFIs
            # if our peer subscribed to a Route Target, it means that we needs
            # to send him all routes of any AFI/SAFI carrying this RouteTarget.
            for (afi, safi) in self._active_families:
                if (afi, safi) != (exa.AFI.ipv4, exa.SAFI.rtc):
                    if action == exa_message.IN.ANNOUNCED:
                        self._subscribe(afi, safi, nlri.rt)
                    elif action == exa_message.IN.WITHDRAWN:
                        self._unsubscribe(afi, safi, nlri.rt)
                    else:
                        raise Exception("unsupported action ??? (%s)" % action)

    def _send(self, data):
        # (error if state not the right one for sending updates)
        self.log.debug("Sending %d bytes on socket to peer %s",
                       len(data), self.peer_address)
        try:
            for _ in self.protocol.connection.writer(data):
                pass
        except Exception:
            self.log.exception("Was not able to send data")

    def _keep_alive_message_data(self):
        return exa_message.KeepAlive().message()

    def _update_for_route_event(self, event):
        try:
            r = exa_message.Update([event.route_entry.nlri],
                                   event.route_entry.attributes)
            return ''.join(r.messages(self.protocol.negotiated))
        except Exception:
            self.log.exception("Exception while generating message for "
                               "route %s", r)
            return ''

    # Looking Glass ###############

    def get_lg_local_info(self, path_prefix):
        return {
            "peeringAddresses": {"peer_address":  self.peer_address,
                                 "local_address": self.local_address},
            "as_info": {"local": cfg.CONF.BGP.my_as,
                        "peer":  cfg.CONF.BGP.my_as},
            "rtc": {"active": self.rtc_active,
                    "enabled": cfg.CONF.BGP.enable_rtc},
            "active_families": [repr(f) for f in self._active_families],
        }



# Copyright (c) 2017 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import uuidutils

EVPN_RT1 = {'import_rt': ['EVPN:1'],
            'export_rt': ['EVPN:1']}

EVPN_RT2 = {'import_rt': ['EVPN:2'],
            'export_rt': ['EVPN:2']}

IPVPN_RT100 = {'import_rt': ['IPVPN:100'],
               'export_rt': ['IPVPN:100']}

IPVPN_RT200 = {'import_rt': ['IPVPN:200'],
               'export_rt': ['IPVPN:200']}

NETWORK_INFO1 = {'network_id': uuidutils.generate_uuid(),
                 'gateway_ip': '10.0.0.1'
                 }

PORT_INFO1 = {'mac_address': '00:00:de:ad:be:ef',
              'ip_address': '10.0.0.2',
              'local_port': {'linuxif': 'port1'},
              }

UPDATED_LOCAL_PORT1 = {'linuxif': 'updated_port1'}

GW_MAC_PORT1 = '00:00:ca:fe:ba:be'

STATIC_ROUTE1 = '1.1.1.1/24'



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Fake BGP Route Reflector

It is a dumb program connecting two TCP sockets together.
It will buffer what is sent by the first TCP client, and send it to
the second one when it arrives.

It can be used to act as a super crude route reflector between
two bagpipe-bgp instances.  But it won't support more than 2 !
"""

import threading

# fakerr is provided as a helper for manuals tests, but we don't want to draw
# twisted as a dependency for the whole package so it is not installed in
# test venvs, so we need that to keep pylint happy:
# pylint: disable=import-error

from twisted.application import internet
from twisted.application import service
from twisted.internet import endpoints
from twisted.internet import protocol
from twisted.internet import reactor
from twisted.protocols import basic

# pylint: disable=no-member


class FakeRR(basic.LineReceiver):

    delimiter = chr(255) * 16

    def __init__(self):
        self.lock = threading.Lock()

    def connectionMade(self):
        with self.lock:
            if self.factory.clients == 2:
                print("Have already 2 peers, not accepting %s, "
                      "reseting everyone !" % self.transport.getPeer().host)
                for client in self.factory.clients:
                    client.transport.loseConnection()
                raise Exception("Have already 2 peers, not accepting more !"
                                "  (%s)" % self.transport.getPeer().host)

            print("Got new peer: %s" % self.transport.getPeer().host)

            self.factory.clients.append(self)
            if len(self.factory.clients) == 2:
                # we are the second client
                print("%s is second peer, sending buffered data..." %
                      self.transport.getPeer().host)
                for olddata in self.factory.buffer:
                    print("   sending buffered data to peer %s (%d bytes)" %
                          (self.transport.getPeer().host, len(olddata)))
                    self.transport.write(olddata)
                self.factory.buffer = []
                self.factory.ready = True
                print("now ready")
            else:
                print("%s is first peer, will buffer data until second peer "
                      "arrives..." % self.transport.getPeer().host)
                self.factory.ready = False
                self.factory.buffer = []

    def connectionLost(self, reason=protocol.Protocol.connectionLost):
        print("Lost peer %s" % self.transport.getPeer().host)
        try:
            self.factory.clients.remove(self)
        except Exception:
            pass

        for c in self.factory.clients:
            if c != self:
                c.transport.loseConnection()
        try:
            self.factory.clients.remove(c)
        except Exception:
            pass

        self.factory.ready = False
        self.factory.buffer = []

    def dataReceived(self, data):
        if self.factory.ready:
            for c in self.factory.clients:
                if c != self:
                    c.transport.write(data)
        else:
            print("buffering received data (%d bytes)" % len(data))
            if self.factory.buffer is None:
                print("??? not ready, but no self.factory.buffer...??")
                self.factory.buffer = []
            self.factory.buffer.append(data)

factory = protocol.ServerFactory()
factory.protocol = FakeRR
factory.clients = []
factory.ready = False
factory.buffer = []

# this application definition allows the use with twisted:
#
#  echo "from networking_bagpipe.bagpipe_bgp.fakerr import application" \
#    | twistd -y /dev/stdin
application = service.Application("fakerr")
internet.TCPServer(179, factory).setServiceParent(application)


def main():
    endpoints.serverFromString(reactor, "tcp:179").listen(factory)
    reactor.run()

if __name__ == '__main__':
    main()

__all__ = ['main']



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

IPVPN = "ipvpn"
EVPN = "evpn"

VPN_TYPES = [EVPN, IPVPN]

RT_IMPORT = 'import_rt'
RT_EXPORT = 'export_rt'
RT_TYPES = [RT_IMPORT, RT_EXPORT]

# port directions
TO_PORT = 'to-port'
FROM_PORT = 'from-port'
BOTH = 'both'
ALL_DIRECTIONS = (BOTH, TO_PORT, FROM_PORT)


def config_group(vpn_type):
    return "DATAPLANE_DRIVER_%s" % vpn_type.upper()

# maximum length for a linux network device name
#  grep 'define.*IFNAMSIZ' /usr/src/linux/include/uapi/linux/if.h
# define    IFNAMSIZ    16
# (minus 1 for trailing null)
LINUX_DEV_LEN = 15



# Copyright (c) 2016 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.common import utils
from neutron.tests.fullstack.resources import client as neutron_client


class BGPVPNClientFixture(neutron_client.ClientFixture):
    """Manage and cleanup BGPVPN resources."""

    def create_bgpvpn(self, tenant_id, name=None, **kwargs):
        resource_type = 'bgpvpn'

        name = name or utils.get_rand_name(prefix='bgpvpn')
        spec = {
            'tenant_id': tenant_id,
            'name': name
        }
        spec.update(kwargs)

        return self._create_resource(resource_type, spec)

    def create_network_association(self, tenant_id, bgpvpn_id, network_id):
        network_association = {
            'network_association': {
                'tenant_id': tenant_id,
                'network_id': network_id
            }
        }
        assoc = self.client.create_network_association(
            bgpvpn_id,
            network_association)
        self.addCleanup(
            neutron_client._safe_method(
                self.client.delete_network_association),
            assoc['network_association']['id'], bgpvpn_id)

    def create_router_association(self, tenant_id, bgpvpn_id, router_id):
        router_association = {
            'router_association': {
                'tenant_id': tenant_id,
                'router_id': router_id
            }
        }
        assoc = self.client.create_router_association(
            bgpvpn_id,
            router_association)
        self.addCleanup(
            neutron_client._safe_method(
                self.client.delete_router_association),
            assoc['router_association']['id'], bgpvpn_id)



# Copyright (c) 2017 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from oslo_serialization import jsonutils
from oslo_utils import importutils

from neutron_lib import context

from neutron.api import extensions as api_ext
from neutron.api.rpc.callbacks import events as rpc_events
from neutron.common import config

from networking_bagpipe.driver import sfc as driver
from networking_bagpipe.objects import sfc as sfc_obj
from networking_bagpipe.tests.unit.driver.sfc import base as sfc_base

from networking_sfc.db import flowclassifier_db as fc_db
from networking_sfc.db import sfc_db
from networking_sfc.extensions import flowclassifier
from networking_sfc.extensions import servicegraph
from networking_sfc.extensions import sfc
from networking_sfc.extensions import tap
from networking_sfc.services.sfc.common import context as sfc_ctx
from networking_sfc.services.sfc.common import exceptions as sfc_exc
from networking_sfc.tests.unit.db import test_flowclassifier_db
from networking_sfc.tests.unit.db import test_sfc_db


class BaGPipeSfcDriverTestCase(
    test_sfc_db.SfcDbPluginTestCaseBase,
    test_flowclassifier_db.FlowClassifierDbPluginTestCaseBase,
    sfc_base.NeutronDbPluginV2TestCase
):

    resource_prefix_map = dict([
        (k, sfc.SFC_PREFIX)
        for k in sfc.RESOURCE_ATTRIBUTE_MAP.keys()
    ] + [
        (k, flowclassifier.FLOW_CLASSIFIER_PREFIX)
        for k in flowclassifier.RESOURCE_ATTRIBUTE_MAP.keys()
    ])

    def setUp(self):
        sfc_plugin = test_sfc_db.DB_SFC_PLUGIN_CLASS
        flowclassifier_plugin = (
            test_flowclassifier_db.DB_FLOWCLASSIFIER_PLUGIN_CLASS)

        service_plugins = {
            sfc.SFC_EXT: sfc_plugin,
            flowclassifier.FLOW_CLASSIFIER_EXT: flowclassifier_plugin
        }
        sfc_db.SfcDbPlugin.supported_extension_aliases = [
            sfc.SFC_EXT, servicegraph.SG_EXT, tap.TAP_EXT]
        sfc_db.SfcDbPlugin.path_prefix = sfc.SFC_PREFIX
        fc_db.FlowClassifierDbPlugin.supported_extension_aliases = [
            flowclassifier.FLOW_CLASSIFIER_EXT]
        fc_db.FlowClassifierDbPlugin.path_prefix = (
            flowclassifier.FLOW_CLASSIFIER_PREFIX
        )
        super(BaGPipeSfcDriverTestCase, self).setUp(
            ext_mgr=None,
            plugin=None,
            service_plugins=service_plugins
        )
        self.sfc_plugin = importutils.import_object(sfc_plugin)
        self.flowclassifier_plugin = importutils.import_object(
            flowclassifier_plugin)
        ext_mgr = api_ext.PluginAwareExtensionManager.get_instance()
        app = config.load_paste_app('extensions_test_app')
        self.ext_api = api_ext.ExtensionMiddleware(app, ext_mgr=ext_mgr)
        self.ctx = context.get_admin_context()

        self.driver = driver.BaGPipeSfcDriver()
        self.driver.initialize()
        self.mocked_push_rpc = self.driver._push_rpc.push = mock.Mock()

    def _validate_chain_hop_rpc_call(self, call_index, rpc_event,
                                     expected_chain_hops):
        rpc_call_args = self.mocked_push_rpc.call_args_list[call_index][0]
        self.assertEqual(len(rpc_call_args[1]), len(expected_chain_hops))
        self.assertEqual(rpc_call_args[2], rpc_event)

        for index, hop in enumerate(rpc_call_args[1]):
            current_hop = hop.to_dict()

            if current_hop.get('classifiers'):
                current_hop['classifiers'] = (
                    jsonutils.loads(current_hop['classifiers'])
                )

            self.assertDictContainsSubset(
                expected_chain_hops[index],
                current_hop)

    def _validate_port_hops_rpc_call(self, call_index, rpc_event,
                                     ingress_port, egress_port):
        rpc_call_args = self.mocked_push_rpc.call_args_list[call_index][0]
        self.assertEqual(len(rpc_call_args[1]), 2)
        self.assertEqual(rpc_call_args[2], rpc_event)

        self.assertEqual(ingress_port, rpc_call_args[1][0].port_id)
        self.assertEqual(len(rpc_call_args[1][0].ingress_hops), 0)
        self.assertEqual(len(rpc_call_args[1][0].egress_hops),
                         int(rpc_event == rpc_events.CREATED))

        if rpc_event == rpc_events.CREATED:
            self.assertIsInstance(rpc_call_args[1][0].egress_hops[0],
                                  sfc_obj.BaGPipeChainHop)

        self.assertEqual(egress_port, rpc_call_args[1][1].port_id)
        self.assertEqual(len(rpc_call_args[1][1].ingress_hops),
                         int(rpc_event == rpc_events.CREATED))
        self.assertEqual(len(rpc_call_args[1][1].egress_hops), 0)

        if rpc_event == rpc_events.CREATED:
            self.assertIsInstance(rpc_call_args[1][1].ingress_hops[0],
                                  sfc_obj.BaGPipeChainHop)

    def test_create_port_chain(self):
        with self.port_pair_group(
            port_pair_group={'name': 'test_ppg'}
        ) as ppg:
            ppg_context = sfc_ctx.PortPairGroupContext(
                self.sfc_plugin, self.ctx,
                ppg['port_pair_group']
            )
            self.driver.create_port_pair_group(ppg_context)
            with self.port_chain(port_chain={
                'name': 'test_chain',
                'port_pair_groups': [
                    ppg['port_pair_group']['id']
                ]
            }) as pc:
                pc_context = sfc_ctx.PortChainContext(
                    self.sfc_plugin, self.ctx,
                    pc['port_chain']
                )
                self.driver.create_port_chain(pc_context)

                self.mocked_push_rpc.assert_not_called()

    def test_create_port_chain_with_port_pairs(self):
        with self.port(
            name='ingress_port',
            device_owner='compute:None',
            device_id='test'
        ) as ingress_port, self.port(
            name='egress_port',
            device_owner='compute:None',
            device_id='test'
        ) as egress_port:
            with self.port_pair(port_pair={
                'ingress': ingress_port['port']['id'],
                'egress': egress_port['port']['id']
            }) as pp:
                pp_context = sfc_ctx.PortPairContext(
                    self.sfc_plugin, self.ctx,
                    pp['port_pair']
                )
                self.driver.create_port_pair(pp_context)
                with self.port_pair_group(port_pair_group={
                    'port_pairs': [pp['port_pair']['id']]
                }) as ppg:
                    ppg_context = sfc_ctx.PortPairGroupContext(
                        self.sfc_plugin, self.ctx,
                        ppg['port_pair_group']
                    )
                    self.driver.create_port_pair_group(ppg_context)
                    with self.port_chain(port_chain={
                        'name': 'test_chain',
                        'port_pair_groups': [
                            ppg['port_pair_group']['id']
                        ]
                    }) as pc:
                        pc_context = sfc_ctx.PortChainContext(
                            self.sfc_plugin, self.ctx,
                            pc['port_chain']
                        )
                        self.driver.create_port_chain(pc_context)

                        self.mocked_push_rpc.assert_not_called()

    def test_create_port_chain_with_flow_classifiers(self):
        with self.port_pair_group(
            port_pair_group={'name': 'test_ppg'}
        ) as ppg:
            ppg_context = sfc_ctx.PortPairGroupContext(
                self.sfc_plugin, self.ctx,
                ppg['port_pair_group']
            )
            self.driver.create_port_pair_group(ppg_context)
            with self.flow_classifier(flow_classifier={
                'source_port_range_min': 100,
                'source_port_range_max': 200,
                'destination_port_range_min': 300,
                'destination_port_range_max': 400,
                'ethertype': 'IPv4',
                'source_ip_prefix': sfc_base.SRC_CIDR,
                'destination_ip_prefix': sfc_base.DEST_CIDR,
                'l7_parameters': {},
                'protocol': 'tcp'
            }) as fc, self.port_chain(port_chain={
                'name': 'test_chain',
                'port_pair_groups': [
                    ppg['port_pair_group']['id']
                ],
                'flow_classifiers': [
                    fc['flow_classifier']['id']
                ]
            }) as pc:
                pc_context = sfc_ctx.PortChainContext(
                    self.sfc_plugin, self.ctx,
                    pc['port_chain']
                )
                self.driver.create_port_chain(pc_context)

                self.mocked_push_rpc.assert_not_called()

    def test_create_port_chain_with_flow_classifiers_port_pairs(self):
        with self.port(
            name='src_port',
            device_owner='compute:None',
            device_id='test',
            network_id=self._src_net['network']['id']
        ) as src_port, self.port(
            name='ingress_port',
            device_owner='compute:None',
            device_id='test'
        ) as ingress_port, self.port(
            name='egress_port',
            device_owner='compute:None',
            device_id='test'
        ) as egress_port, self.port(
            name='dest_port',
            device_owner='compute:None',
            device_id='test',
            network_id=self._dest_net['network']['id']
        ) as dest_port:
            with self.port_pair(port_pair={
                'ingress': ingress_port['port']['id'],
                'egress': egress_port['port']['id']
            }) as pp:
                pp_context = sfc_ctx.PortPairContext(
                    self.sfc_plugin, self.ctx,
                    pp['port_pair']
                )
                self.driver.create_port_pair(pp_context)
                with self.port_pair_group(port_pair_group={
                    'port_pairs': [
                        pp['port_pair']['id']
                    ]
                }) as ppg:
                    ppg_context = sfc_ctx.PortPairGroupContext(
                        self.sfc_plugin, self.ctx,
                        ppg['port_pair_group']
                    )
                    self.driver.create_port_pair_group(ppg_context)
                    with self.flow_classifier(flow_classifier={
                        'source_port_range_min': 100,
                        'source_port_range_max': 200,
                        'destination_port_range_min': 300,
                        'destination_port_range_max': 400,
                        'ethertype': 'IPv4',
                        'source_ip_prefix': sfc_base.SRC_CIDR,
                        'destination_ip_prefix': sfc_base.DEST_CIDR,
                        'l7_parameters': {},
                        'protocol': 'tcp',
                        'logical_source_port': src_port['port']['id']
                    }) as fc, self.port_chain(port_chain={
                        'name': 'test_chain',
                        'port_pair_groups': [
                            ppg['port_pair_group']['id']
                        ],
                        'flow_classifiers': [
                            fc['flow_classifier']['id']
                        ]
                    }) as pc:
                        pc_context = sfc_ctx.PortChainContext(
                            self.sfc_plugin, self.ctx,
                            pc['port_chain']
                        )
                        self.driver.create_port_chain(pc_context)

                        expected_chain_hops = [dict(
                            project_id=self._tenant_id,
                            portchain_id=pc['port_chain']['id'],
                            ingress_ppg=ppg['port_pair_group']['id'],
                            egress_ppg=None,
                            ingress_network=None,
                            egress_network=self._dest_net['network']['id'],
                            ingress_gw=self._subnet['subnet']['gateway_ip'],
                            egress_gw=sfc_base.DEST_GATEWAY,
                            ingress_ports=[egress_port['port']['id']],
                            egress_ports=[dest_port['port']['id']],
                            rts=[self._dest_rt],
                            readv_from_rts=[],
                            readv_to_rt=None,
                            redirect_rts=[],
                            classifiers=None,
                            reverse_hop=False
                        ), dict(
                            project_id=self._tenant_id,
                            portchain_id=pc['port_chain']['id'],
                            ingress_ppg=None,
                            egress_ppg=ppg['port_pair_group']['id'],
                            ingress_network=self._src_net['network']['id'],
                            egress_network=None,
                            ingress_gw=sfc_base.SRC_GATEWAY,
                            egress_gw=self._subnet['subnet']['gateway_ip'],
                            ingress_ports=[src_port['port']['id']],
                            egress_ports=[ingress_port['port']['id']],
                            rts=['64512:5000'],
                            readv_from_rts=[],
                            readv_to_rt=None,
                            redirect_rts=['64512:5000'],
                            attract_to_rt='64512:5001',
                            classifiers=[
                                dict(protocol='tcp',
                                     sourcePrefix=sfc_base.SRC_CIDR,
                                     sourcePort='100:200',
                                     destinationPrefix=sfc_base.DEST_CIDR,
                                     destinationPort='300:400')],
                            reverse_hop=False
                        )]

                        self.assertEqual(self.mocked_push_rpc.call_count, 1)
                        self._validate_chain_hop_rpc_call(0,
                                                          rpc_events.CREATED,
                                                          expected_chain_hops)

    def test_create_port_chain_multi_port_pair_groups_port_pairs(self):
        with self.port(
            name='src_port',
            device_owner='compute:None',
            device_id='test',
            network_id=self._src_net['network']['id']
        ) as src_port, self.port(
            name='ingress_port1',
            device_owner='compute:None',
            device_id='test'
        ) as ingress_port1, self.port(
            name='egress_port1',
            device_owner='compute:None',
            device_id='test'
        ) as egress_port1, self.port(
            name='ingress_port2',
            device_owner='compute:None',
            device_id='test'
        ) as ingress_port2, self.port(
            name='egress_port2',
            device_owner='compute:None',
            device_id='test'
        ) as egress_port2, self.port(
            name='ingress_port3',
            device_owner='compute:None',
            device_id='test'
        ) as ingress_port3, self.port(
            name='egress_port3',
            device_owner='compute:None',
            device_id='test'
        ) as egress_port3, self.port(
            name='dest_port',
            device_owner='compute:None',
            device_id='test',
            network_id=self._dest_net['network']['id']
        ) as dest_port:
            with self.port_pair(port_pair={
                'ingress': ingress_port1['port']['id'],
                'egress': egress_port1['port']['id']
            }) as pp1, self.port_pair(port_pair={
                'ingress': ingress_port2['port']['id'],
                'egress': egress_port2['port']['id']
            }) as pp2, self.port_pair(port_pair={
                'ingress': ingress_port3['port']['id'],
                'egress': egress_port3['port']['id']
            }) as pp3:
                pp1_context = sfc_ctx.PortPairContext(
                    self.sfc_plugin, self.ctx,
                    pp1['port_pair']
                )
                self.driver.create_port_pair(pp1_context)
                pp2_context = sfc_ctx.PortPairContext(
                    self.sfc_plugin, self.ctx,
                    pp2['port_pair']
                )
                self.driver.create_port_pair(pp2_context)
                pp3_context = sfc_ctx.PortPairContext(
                    self.sfc_plugin, self.ctx,
                    pp3['port_pair']
                )
                self.driver.create_port_pair(pp3_context)
                with self.port_pair_group(port_pair_group={
                    'port_pairs': [
                        pp1['port_pair']['id']
                    ]
                }) as ppg1, self.port_pair_group(port_pair_group={
                    'port_pairs': [
                        pp2['port_pair']['id']
                    ]
                }) as ppg2, self.port_pair_group(port_pair_group={
                    'port_pairs': [
                        pp3['port_pair']['id']
                    ]
                }) as ppg3:
                    ppg1_context = sfc_ctx.PortPairGroupContext(
                        self.sfc_plugin, self.ctx,
                        ppg1['port_pair_group']
                    )
                    self.driver.create_port_pair_group(ppg1_context)
                    ppg2_context = sfc_ctx.PortPairGroupContext(
                        self.sfc_plugin, self.ctx,
                        ppg2['port_pair_group']
                    )
                    self.driver.create_port_pair_group(ppg2_context)
                    ppg3_context = sfc_ctx.PortPairGroupContext(
                        self.sfc_plugin, self.ctx,
                        ppg3['port_pair_group']
                    )
                    self.driver.create_port_pair_group(ppg3_context)
                    with self.flow_classifier(flow_classifier={
                        'source_port_range_min': 100,
                        'source_port_range_max': 200,
                        'destination_port_range_min': 300,
                        'destination_port_range_max': 400,
                        'ethertype': 'IPv4',
                        'source_ip_prefix': sfc_base.SRC_CIDR,
                        'destination_ip_prefix': sfc_base.DEST_CIDR,
                        'l7_parameters': {},
                        'protocol': 'tcp',
                        'logical_source_port': src_port['port']['id']
                    }) as fc, self.port_chain(port_chain={
                        'name': 'test_chain',
                        'port_pair_groups': [
                            ppg1['port_pair_group']['id'],
                            ppg2['port_pair_group']['id'],
                            ppg3['port_pair_group']['id']
                        ],
                        'flow_classifiers': [
                            fc['flow_classifier']['id']
                        ]
                    }) as pc:
                        pc_context = sfc_ctx.PortChainContext(
                            self.sfc_plugin, self.ctx,
                            pc['port_chain']
                        )
                        self.driver.create_port_chain(pc_context)

                        expected_chain_hops = [dict(
                            project_id=self._tenant_id,
                            portchain_id=pc['port_chain']['id'],
                            ingress_ppg=ppg3['port_pair_group']['id'],
                            egress_ppg=None,
                            ingress_network=None,
                            egress_network=self._dest_net['network']['id'],
                            ingress_gw=self._subnet['subnet']['gateway_ip'],
                            egress_gw=sfc_base.DEST_GATEWAY,
                            ingress_ports=[egress_port3['port']['id']],
                            egress_ports=[dest_port['port']['id']],
                            rts=[self._dest_rt],
                            readv_from_rts=[],
                            readv_to_rt=None,
                            redirect_rts=[],
                            classifiers=None,
                            reverse_hop=False
                        ), dict(
                            project_id=self._tenant_id,
                            portchain_id=pc['port_chain']['id'],
                            ingress_ppg=ppg2['port_pair_group']['id'],
                            egress_ppg=ppg3['port_pair_group']['id'],
                            ingress_network=None,
                            egress_network=None,
                            ingress_gw=self._subnet['subnet']['gateway_ip'],
                            egress_gw=self._subnet['subnet']['gateway_ip'],
                            ingress_ports=[egress_port2['port']['id']],
                            egress_ports=[ingress_port3['port']['id']],
                            rts=['64512:5000'],
                            readv_from_rts=[],
                            readv_to_rt=None,
                            redirect_rts=['64512:5000'],
                            attract_to_rt='64512:5001',
                            classifiers=[
                                dict(protocol='tcp',
                                     sourcePrefix=sfc_base.SRC_CIDR,
                                     sourcePort='100:200',
                                     destinationPrefix=sfc_base.DEST_CIDR,
                                     destinationPort='300:400')],
                            reverse_hop=False
                        ), dict(
                            project_id=self._tenant_id,
                            portchain_id=pc['port_chain']['id'],
                            ingress_ppg=ppg1['port_pair_group']['id'],
                            egress_ppg=ppg2['port_pair_group']['id'],
                            ingress_network=None,
                            egress_network=None,
                            ingress_gw=self._subnet['subnet']['gateway_ip'],
                            egress_gw=self._subnet['subnet']['gateway_ip'],
                            ingress_ports=[egress_port1['port']['id']],
                            egress_ports=[ingress_port2['port']['id']],
                            rts=['64512:5002'],
                            readv_from_rts=[],
                            readv_to_rt=None,
                            redirect_rts=['64512:5002'],
                            attract_to_rt='64512:5003',
                            classifiers=[
                                dict(protocol='tcp',
                                     sourcePrefix=sfc_base.SRC_CIDR,
                                     sourcePort='100:200',
                                     destinationPrefix=sfc_base.DEST_CIDR,
                                     destinationPort='300:400')],
                            reverse_hop=False
                        ), dict(
                            project_id=self._tenant_id,
                            portchain_id=pc['port_chain']['id'],
                            ingress_ppg=None,
                            egress_ppg=ppg1['port_pair_group']['id'],
                            ingress_network=self._src_net['network']['id'],
                            egress_network=None,
                            ingress_gw=sfc_base.SRC_GATEWAY,
                            egress_gw=self._subnet['subnet']['gateway_ip'],
                            ingress_ports=[src_port['port']['id']],
                            egress_ports=[ingress_port1['port']['id']],
                            rts=['64512:5004'],
                            readv_from_rts=[],
                            readv_to_rt=None,
                            redirect_rts=['64512:5004'],
                            attract_to_rt='64512:5005',
                            classifiers=[
                                dict(protocol='tcp',
                                     sourcePrefix=sfc_base.SRC_CIDR,
                                     sourcePort='100:200',
                                     destinationPrefix=sfc_base.DEST_CIDR,
                                     destinationPort='300:400')],
                            reverse_hop=False
                        )]

                        self.assertEqual(self.mocked_push_rpc.call_count, 1)
                        self._validate_chain_hop_rpc_call(0,
                                                          rpc_events.CREATED,
                                                          expected_chain_hops)

    def test_create_port_chain_multi_port_pair_groups_multi_port_pairs(self):
        with self.port(
            name='src_port',
            device_owner='compute:None',
            device_id='test',
            network_id=self._src_net['network']['id']
        ) as src_port, self.port(
            name='ingress_port1',
            device_owner='compute:None',
            device_id='test'
        ) as ingress_port1, self.port(
            name='egress_port1',
            device_owner='compute:None',
            device_id='test'
        ) as egress_port1, self.port(
            name='ingress_port2',
            device_owner='compute:None',
            device_id='test'
        ) as ingress_port2, self.port(
            name='egress_port2',
            device_owner='compute:None',
            device_id='test'
        ) as egress_port2, self.port(
            name='ingress_port3',
            device_owner='compute:None',
            device_id='test'
        ) as ingress_port3, self.port(
            name='egress_port3',
            device_owner='compute:None',
            device_id='test'
        ) as egress_port3, self.port(
            name='ingress_port4',
            device_owner='compute:None',
            device_id='test'
        ) as ingress_port4, self.port(
            name='egress_port4',
            device_owner='compute:None',
            device_id='test'
        ) as egress_port4, self.port(
            name='dest_port',
            device_owner='compute:None',
            device_id='test',
            network_id=self._dest_net['network']['id']
        ) as dest_port:
            with self.port_pair(port_pair={
                'ingress': ingress_port1['port']['id'],
                'egress': egress_port1['port']['id']
            }) as pp1, self.port_pair(port_pair={
                'ingress': ingress_port2['port']['id'],
                'egress': egress_port2['port']['id']
            }) as pp2, self.port_pair(port_pair={
                'ingress': ingress_port3['port']['id'],
                'egress': egress_port3['port']['id']
            }) as pp3, self.port_pair(port_pair={
                'ingress': ingress_port4['port']['id'],
                'egress': egress_port4['port']['id']
            }) as pp4:
                pp1_context = sfc_ctx.PortPairContext(
                    self.sfc_plugin, self.ctx,
                    pp1['port_pair']
                )
                self.driver.create_port_pair(pp1_context)
                pp2_context = sfc_ctx.PortPairContext(
                    self.sfc_plugin, self.ctx,
                    pp2['port_pair']
                )
                self.driver.create_port_pair(pp2_context)
                pp3_context = sfc_ctx.PortPairContext(
                    self.sfc_plugin, self.ctx,
                    pp3['port_pair']
                )
                self.driver.create_port_pair(pp3_context)
                pp4_context = sfc_ctx.PortPairContext(
                    self.sfc_plugin, self.ctx,
                    pp4['port_pair']
                )
                self.driver.create_port_pair(pp4_context)
                with self.port_pair_group(port_pair_group={
                    'port_pairs': [
                        pp1['port_pair']['id'],
                        pp2['port_pair']['id']
                    ]
                }) as ppg1, self.port_pair_group(port_pair_group={
                    'port_pairs': [
                        pp3['port_pair']['id'],
                        pp4['port_pair']['id']
                    ]
                }) as ppg2:
                    ppg1_context = sfc_ctx.PortPairGroupContext(
                        self.sfc_plugin, self.ctx,
                        ppg1['port_pair_group']
                    )
                    self.driver.create_port_pair_group(ppg1_context)
                    ppg2_context = sfc_ctx.PortPairGroupContext(
                        self.sfc_plugin, self.ctx,
                        ppg2['port_pair_group']
                    )
                    self.driver.create_port_pair_group(ppg2_context)
                    with self.flow_classifier(flow_classifier={
                        'source_port_range_min': 100,
                        'source_port_range_max': 200,
                        'destination_port_range_min': 300,
                        'destination_port_range_max': 400,
                        'ethertype': 'IPv4',
                        'source_ip_prefix': sfc_base.SRC_CIDR,
                        'destination_ip_prefix': sfc_base.DEST_CIDR,
                        'l7_parameters': {},
                        'protocol': 'tcp',
                        'logical_source_port': src_port['port']['id']
                    }) as fc, self.port_chain(port_chain={
                        'name': 'test_chain',
                        'port_pair_groups': [
                            ppg1['port_pair_group']['id'],
                            ppg2['port_pair_group']['id']
                        ],
                        'flow_classifiers': [
                            fc['flow_classifier']['id']
                        ]
                    }) as pc:
                        pc_context = sfc_ctx.PortChainContext(
                            self.sfc_plugin, self.ctx,
                            pc['port_chain']
                        )
                        self.driver.create_port_chain(pc_context)

                        expected_chain_hops = [dict(
                            project_id=self._tenant_id,
                            portchain_id=pc['port_chain']['id'],
                            ingress_ppg=ppg2['port_pair_group']['id'],
                            egress_ppg=None,
                            ingress_network=None,
                            egress_network=self._dest_net['network']['id'],
                            ingress_gw=self._subnet['subnet']['gateway_ip'],
                            egress_gw=sfc_base.DEST_GATEWAY,
                            ingress_ports=[egress_port3['port']['id'],
                                           egress_port4['port']['id']],
                            egress_ports=[dest_port['port']['id']],
                            rts=[self._dest_rt],
                            readv_from_rts=[],
                            readv_to_rt=None,
                            redirect_rts=[],
                            classifiers=None,
                            reverse_hop=False
                        ), dict(
                            project_id=self._tenant_id,
                            portchain_id=pc['port_chain']['id'],
                            ingress_ppg=ppg1['port_pair_group']['id'],
                            egress_ppg=ppg2['port_pair_group']['id'],
                            ingress_network=None,
                            egress_network=None,
                            ingress_gw=self._subnet['subnet']['gateway_ip'],
                            egress_gw=self._subnet['subnet']['gateway_ip'],
                            ingress_ports=[egress_port1['port']['id'],
                                           egress_port2['port']['id']],
                            egress_ports=[ingress_port3['port']['id'],
                                          ingress_port4['port']['id']],
                            rts=['64512:5000'],
                            readv_from_rts=[],
                            readv_to_rt=None,
                            redirect_rts=['64512:5000'],
                            attract_to_rt='64512:5001',
                            classifiers=[
                                dict(protocol='tcp',
                                     sourcePrefix=sfc_base.SRC_CIDR,
                                     sourcePort='100:200',
                                     destinationPrefix=sfc_base.DEST_CIDR,
                                     destinationPort='300:400')],
                            reverse_hop=False
                        ), dict(
                            project_id=self._tenant_id,
                            portchain_id=pc['port_chain']['id'],
                            ingress_ppg=None,
                            egress_ppg=ppg1['port_pair_group']['id'],
                            ingress_network=self._src_net['network']['id'],
                            egress_network=None,
                            ingress_gw=sfc_base.SRC_GATEWAY,
                            egress_gw=self._subnet['subnet']['gateway_ip'],
                            ingress_ports=[src_port['port']['id']],
                            egress_ports=[ingress_port1['port']['id'],
                                          ingress_port2['port']['id']],
                            rts=['64512:5002'],
                            readv_from_rts=[],
                            readv_to_rt=None,
                            redirect_rts=['64512:5002'],
                            attract_to_rt='64512:5003',
                            classifiers=[
                                dict(protocol='tcp',
                                     sourcePrefix=sfc_base.SRC_CIDR,
                                     sourcePort='100:200',
                                     destinationPrefix=sfc_base.DEST_CIDR,
                                     destinationPort='300:400')],
                            reverse_hop=False
                        )]

                        self.assertEqual(self.mocked_push_rpc.call_count, 1)
                        self._validate_chain_hop_rpc_call(0,
                                                          rpc_events.CREATED,
                                                          expected_chain_hops)

    def test_create_port_chain_with_symmetric_parameter(self):
        with self.port(
            name='src_port',
            device_owner='compute:None',
            device_id='test',
            network_id=self._src_net['network']['id']
        ) as src_port, self.port(
            name='ingress_port',
            device_owner='compute:None',
            device_id='test'
        ) as ingress_port, self.port(
            name='egress_port',
            device_owner='compute:None',
            device_id='test'
        ) as egress_port, self.port(
            name='dest_port',
            device_owner='compute:None',
            device_id='test',
            network_id=self._dest_net['network']['id']
        ) as dest_port:
            with self.port_pair(port_pair={
                'ingress': ingress_port['port']['id'],
                'egress': egress_port['port']['id']
            }) as pp:
                pp_context = sfc_ctx.PortPairContext(
                    self.sfc_plugin, self.ctx,
                    pp['port_pair']
                )
                self.driver.create_port_pair(pp_context)
                with self.port_pair_group(port_pair_group={
                    'port_pairs': [
                        pp['port_pair']['id']
                    ]
                }) as ppg:
                    ppg_context = sfc_ctx.PortPairGroupContext(
                        self.sfc_plugin, self.ctx,
                        ppg['port_pair_group']
                    )
                    self.driver.create_port_pair_group(ppg_context)
                    with self.flow_classifier(flow_classifier={
                        'source_port_range_min': 100,
                        'source_port_range_max': 200,
                        'destination_port_range_min': 300,
                        'destination_port_range_max': 400,
                        'ethertype': 'IPv4',
                        'source_ip_prefix': sfc_base.SRC_CIDR,
                        'destination_ip_prefix': sfc_base.DEST_CIDR,
                        'l7_parameters': {},
                        'protocol': 'tcp',
                        'logical_source_port': src_port['port']['id']
                    }) as fc, self.port_chain(port_chain={
                        'name': 'test_chain',
                        'port_pair_groups': [
                            ppg['port_pair_group']['id']
                        ],
                        'flow_classifiers': [
                            fc['flow_classifier']['id']
                        ],
                        'chain_parameters': {'symmetric': True}
                    }) as pc:
                        pc_context = sfc_ctx.PortChainContext(
                            self.sfc_plugin, self.ctx,
                            pc['port_chain']
                        )
                        self.driver.create_port_chain(pc_context)

                        expected_chain_hops = [dict(
                            project_id=self._tenant_id,
                            portchain_id=pc['port_chain']['id'],
                            ingress_ppg=ppg['port_pair_group']['id'],
                            egress_ppg=None,
                            ingress_network=None,
                            egress_network=self._dest_net['network']['id'],
                            ingress_gw=self._subnet['subnet']['gateway_ip'],
                            egress_gw=sfc_base.DEST_GATEWAY,
                            ingress_ports=[egress_port['port']['id']],
                            egress_ports=[dest_port['port']['id']],
                            rts=[self._dest_rt],
                            readv_from_rts=[],
                            readv_to_rt=None,
                            redirect_rts=[],
                            classifiers=None,
                            reverse_hop=False
                        ), dict(
                            project_id=self._tenant_id,
                            portchain_id=pc['port_chain']['id'],
                            ingress_ppg=None,
                            egress_ppg=ppg['port_pair_group']['id'],
                            ingress_network=self._src_net['network']['id'],
                            egress_network=None,
                            ingress_gw=sfc_base.SRC_GATEWAY,
                            egress_gw=self._subnet['subnet']['gateway_ip'],
                            ingress_ports=[src_port['port']['id']],
                            egress_ports=[ingress_port['port']['id']],
                            rts=['64512:5000'],
                            readv_from_rts=[],
                            readv_to_rt=None,
                            redirect_rts=['64512:5000'],
                            attract_to_rt='64512:5001',
                            classifiers=[
                                dict(protocol='tcp',
                                     sourcePrefix=sfc_base.SRC_CIDR,
                                     sourcePort='100:200',
                                     destinationPrefix=sfc_base.DEST_CIDR,
                                     destinationPort='300:400')],
                            reverse_hop=False
                        ), dict(
                            project_id=self._tenant_id,
                            portchain_id=pc['port_chain']['id'],
                            ingress_ppg=ppg['port_pair_group']['id'],
                            egress_ppg=None,
                            ingress_network=None,
                            egress_network=self._src_net['network']['id'],
                            ingress_gw=self._subnet['subnet']['gateway_ip'],
                            egress_gw=sfc_base.SRC_GATEWAY,
                            ingress_ports=[ingress_port['port']['id']],
                            egress_ports=[src_port['port']['id']],
                            rts=[self._src_rt],
                            readv_from_rts=[],
                            readv_to_rt=None,
                            redirect_rts=[],
                            classifiers=None,
                            reverse_hop=True
                        ), dict(
                            project_id=self._tenant_id,
                            portchain_id=pc['port_chain']['id'],
                            ingress_ppg=None,
                            egress_ppg=ppg['port_pair_group']['id'],
                            ingress_network=self._dest_net['network']['id'],
                            egress_network=None,
                            ingress_gw=sfc_base.DEST_GATEWAY,
                            egress_gw=self._subnet['subnet']['gateway_ip'],
                            ingress_ports=[dest_port['port']['id']],
                            egress_ports=[egress_port['port']['id']],
                            rts=['64512:5002'],
                            readv_from_rts=[],
                            readv_to_rt=None,
                            redirect_rts=['64512:5002'],
                            attract_to_rt='64512:5003',
                            classifiers=[
                                dict(protocol='tcp',
                                     sourcePrefix=sfc_base.DEST_CIDR,
                                     sourcePort='300:400',
                                     destinationPrefix=sfc_base.SRC_CIDR,
                                     destinationPort='100:200')],
                            reverse_hop=True
                        )]

                        self.assertEqual(self.mocked_push_rpc.call_count, 1)
                        self._validate_chain_hop_rpc_call(0,
                                                          rpc_events.CREATED,
                                                          expected_chain_hops)

    def test_update_port_chain_add_port_pair(self):
        with self.port(
            name='src_port',
            device_owner='compute:None',
            device_id='test',
            network_id=self._src_net['network']['id']
        ) as src_port, self.port(
            name='ingress_port1',
            device_owner='compute:None',
            device_id='test'
        ) as ingress_port1, self.port(
            name='egress_port1',
            device_owner='compute:None',
            device_id='test'
        ) as egress_port1, self.port(
            name='ingress_port2',
            device_owner='compute:None',
            device_id='test'
        ) as ingress_port2, self.port(
            name='egress_port2',
            device_owner='compute:None',
            device_id='test'
        ) as egress_port2:
            with self.port_pair(port_pair={
                'ingress': ingress_port1['port']['id'],
                'egress': egress_port1['port']['id']
            }) as pp1, self.port_pair(port_pair={
                'ingress': ingress_port2['port']['id'],
                'egress': egress_port2['port']['id']
            }) as pp2:
                pp1_context = sfc_ctx.PortPairContext(
                    self.sfc_plugin, self.ctx,
                    pp1['port_pair']
                )
                self.driver.create_port_pair(pp1_context)
                pp2_context = sfc_ctx.PortPairContext(
                    self.sfc_plugin, self.ctx,
                    pp2['port_pair']
                )
                self.driver.create_port_pair(pp2_context)

                with self.port_pair_group(port_pair_group={
                    'port_pairs': [
                        pp1['port_pair']['id']
                    ]
                }) as ppg:
                    ppg_context = sfc_ctx.PortPairGroupContext(
                        self.sfc_plugin, self.ctx,
                        ppg['port_pair_group']
                    )
                    self.driver.create_port_pair_group(ppg_context)

                    with self.flow_classifier(flow_classifier={
                        'logical_source_port': src_port['port']['id'],
                        'destination_ip_prefix': sfc_base.DEST_CIDR,
                    }) as fc, self.port_chain(port_chain={
                        'name': 'test_chain',
                        'port_pair_groups': [
                            ppg['port_pair_group']['id']
                        ],
                        'flow_classifiers': [
                            fc['flow_classifier']['id']
                        ]
                    }) as pc:
                        pc_context = sfc_ctx.PortChainContext(
                            self.sfc_plugin, self.ctx,
                            pc['port_chain']
                        )
                        self.driver.create_port_chain(pc_context)

                        self.mocked_push_rpc.reset_mock()

                        updates = {
                            'port_pairs': [
                                pp1['port_pair']['id'],
                                pp2['port_pair']['id']
                            ]
                        }
                        req = self.new_update_request(
                            'port_pair_groups', {'port_pair_group': updates},
                            ppg['port_pair_group']['id']
                        )
                        res = req.get_response(self.ext_api)
                        ppgbis = self.deserialize(
                            self.fmt, res
                        )
                        ppgbis['port_pair_group']['port_chains'] = [
                            pc['port_chain']['id']
                        ]
                        ppgbis_context = sfc_ctx.PortPairGroupContext(
                            self.sfc_plugin, self.ctx,
                            ppgbis['port_pair_group'],
                            original_portpairgroup=ppg['port_pair_group']
                        )
                        self.driver.update_port_pair_group(ppgbis_context)

                        self.assertEqual(self.mocked_push_rpc.call_count, 1)
                        self._validate_port_hops_rpc_call(
                            0,
                            rpc_events.CREATED,
                            ingress_port2['port']['id'],
                            egress_port2['port']['id']
                        )

    def test_update_port_chain_delete_port_pair(self):
        with self.port(
            name='src_port',
            device_owner='compute:None',
            device_id='test',
            network_id=self._src_net['network']['id']
        ) as src_port, self.port(
            name='ingress_port1',
            device_owner='compute:None',
            device_id='test'
        ) as ingress_port1, self.port(
            name='egress_port1',
            device_owner='compute:None',
            device_id='test'
        ) as egress_port1, self.port(
            name='ingress_port2',
            device_owner='compute:None',
            device_id='test'
        ) as ingress_port2, self.port(
            name='egress_port2',
            device_owner='compute:None',
            device_id='test'
        ) as egress_port2:
            with self.port_pair(port_pair={
                'ingress': ingress_port1['port']['id'],
                'egress': egress_port1['port']['id']
            }) as pp1, self.port_pair(port_pair={
                'ingress': ingress_port2['port']['id'],
                'egress': egress_port2['port']['id']
            }) as pp2:
                pp1_context = sfc_ctx.PortPairContext(
                    self.sfc_plugin, self.ctx,
                    pp1['port_pair']
                )
                self.driver.create_port_pair(pp1_context)
                pp2_context = sfc_ctx.PortPairContext(
                    self.sfc_plugin, self.ctx,
                    pp2['port_pair']
                )
                self.driver.create_port_pair(pp2_context)

                with self.port_pair_group(port_pair_group={
                    'port_pairs': [
                        pp1['port_pair']['id'],
                        pp2['port_pair']['id']
                    ]
                }) as ppg:
                    ppg_context = sfc_ctx.PortPairGroupContext(
                        self.sfc_plugin, self.ctx,
                        ppg['port_pair_group']
                    )
                    self.driver.create_port_pair_group(ppg_context)

                    with self.flow_classifier(flow_classifier={
                        'logical_source_port': src_port['port']['id'],
                        'destination_ip_prefix': sfc_base.DEST_CIDR,
                    }) as fc, self.port_chain(port_chain={
                        'name': 'test_chain',
                        'port_pair_groups': [
                            ppg['port_pair_group']['id']
                        ],
                        'flow_classifiers': [
                            fc['flow_classifier']['id']
                        ]
                    }) as pc:
                        pc_context = sfc_ctx.PortChainContext(
                            self.sfc_plugin, self.ctx,
                            pc['port_chain']
                        )
                        self.driver.create_port_chain(pc_context)

                        self.mocked_push_rpc.reset_mock()

                        updates = {
                            'port_pairs': [
                                pp1['port_pair']['id']
                            ]
                        }
                        req = self.new_update_request(
                            'port_pair_groups', {'port_pair_group': updates},
                            ppg['port_pair_group']['id']
                        )
                        res = req.get_response(self.ext_api)
                        ppgbis = self.deserialize(
                            self.fmt, res
                        )
                        ppgbis['port_pair_group']['port_chains'] = [
                            pc['port_chain']['id']
                        ]
                        ppgbis_context = sfc_ctx.PortPairGroupContext(
                            self.sfc_plugin, self.ctx,
                            ppgbis['port_pair_group'],
                            original_portpairgroup=ppg['port_pair_group']
                        )
                        self.driver.update_port_pair_group(ppgbis_context)

                        self.assertEqual(self.mocked_push_rpc.call_count, 1)
                        self._validate_port_hops_rpc_call(
                            0,
                            rpc_events.DELETED,
                            ingress_port2['port']['id'],
                            egress_port2['port']['id']
                        )

    def test_update_port_chain_replace_port_pair(self):
        with self.port(
            name='src_port',
            device_owner='compute:None',
            device_id='test',
            network_id=self._src_net['network']['id']
        ) as src_port, self.port(
            name='ingress_port1',
            device_owner='compute:None',
            device_id='test'
        ) as ingress_port1, self.port(
            name='egress_port1',
            device_owner='compute:None',
            device_id='test'
        ) as egress_port1, self.port(
            name='ingress_port2',
            device_owner='compute:None',
            device_id='test'
        ) as ingress_port2, self.port(
            name='egress_port2',
            device_owner='compute:None',
            device_id='test'
        ) as egress_port2:
            with self.port_pair(port_pair={
                'ingress': ingress_port1['port']['id'],
                'egress': egress_port1['port']['id']
            }) as pp1, self.port_pair(port_pair={
                'ingress': ingress_port2['port']['id'],
                'egress': egress_port2['port']['id']
            }) as pp2:
                pp1_context = sfc_ctx.PortPairContext(
                    self.sfc_plugin, self.ctx,
                    pp1['port_pair']
                )
                self.driver.create_port_pair(pp1_context)
                pp2_context = sfc_ctx.PortPairContext(
                    self.sfc_plugin, self.ctx,
                    pp2['port_pair']
                )
                self.driver.create_port_pair(pp2_context)

                with self.port_pair_group(port_pair_group={
                    'port_pairs': [
                        pp1['port_pair']['id']
                    ]
                }) as ppg:
                    ppg_context = sfc_ctx.PortPairGroupContext(
                        self.sfc_plugin, self.ctx,
                        ppg['port_pair_group']
                    )
                    self.driver.create_port_pair_group(ppg_context)

                    with self.flow_classifier(flow_classifier={
                        'logical_source_port': src_port['port']['id'],
                        'destination_ip_prefix': sfc_base.DEST_CIDR
                    }) as fc, self.port_chain(port_chain={
                        'name': 'test_chain',
                        'port_pair_groups': [
                            ppg['port_pair_group']['id']
                        ],
                        'flow_classifiers': [
                            fc['flow_classifier']['id']
                        ]
                    }) as pc:
                        pc_context = sfc_ctx.PortChainContext(
                            self.sfc_plugin, self.ctx,
                            pc['port_chain']
                        )
                        self.driver.create_port_chain(pc_context)

                        self.mocked_push_rpc.reset_mock()

                        updates = {
                            'port_pairs': [
                                pp2['port_pair']['id']
                            ]
                        }
                        req = self.new_update_request(
                            'port_pair_groups', {'port_pair_group': updates},
                            ppg['port_pair_group']['id']
                        )
                        res = req.get_response(self.ext_api)
                        ppgbis = self.deserialize(
                            self.fmt, res
                        )
                        ppgbis['port_pair_group']['port_chains'] = [
                            pc['port_chain']['id']
                        ]
                        ppgbis_context = sfc_ctx.PortPairGroupContext(
                            self.sfc_plugin, self.ctx,
                            ppgbis['port_pair_group'],
                            original_portpairgroup=ppg['port_pair_group']
                        )
                        self.driver.update_port_pair_group(ppgbis_context)

                        self.assertEqual(self.mocked_push_rpc.call_count, 2)
                        self._validate_port_hops_rpc_call(
                            0,
                            rpc_events.CREATED,
                            ingress_port2['port']['id'],
                            egress_port2['port']['id']
                        )

                        self._validate_port_hops_rpc_call(
                            1,
                            rpc_events.DELETED,
                            ingress_port1['port']['id'],
                            egress_port1['port']['id']
                        )

    def test_update_port_chain_delete_flow_classifier(self):
        with self.port(
            name='src_port1',
            device_owner='compute:None',
            device_id='test',
            network_id=self._src_net['network']['id']
        ) as src_port, self.port(
            name='ingress_port',
            device_owner='compute:None',
            device_id='test'
        ) as ingress_port, self.port(
            name='egress_port',
            device_owner='compute:None',
            device_id='test'
        ) as egress_port, self.port(
            name='dest_port',
            device_owner='compute:None',
            device_id='test',
            network_id=self._dest_net['network']['id']
        ) as dest_port:
            with self.port_pair(port_pair={
                'ingress': ingress_port['port']['id'],
                'egress': egress_port['port']['id']
            }) as pp:
                pp_context = sfc_ctx.PortPairContext(
                    self.sfc_plugin, self.ctx,
                    pp['port_pair']
                )
                self.driver.create_port_pair(pp_context)

                with self.port_pair_group(port_pair_group={
                    'port_pairs': [
                        pp['port_pair']['id']
                    ]
                }) as ppg:
                    ppg_context = sfc_ctx.PortPairGroupContext(
                        self.sfc_plugin, self.ctx,
                        ppg['port_pair_group']
                    )
                    self.driver.create_port_pair_group(ppg_context)

                    with self.flow_classifier(flow_classifier={
                        'source_port_range_min': 100,
                        'source_port_range_max': 200,
                        'destination_port_range_min': 300,
                        'destination_port_range_max': 400,
                        'ethertype': 'IPv4',
                        'source_ip_prefix': sfc_base.SRC_CIDR,
                        'destination_ip_prefix': sfc_base.DEST_CIDR,
                        'l7_parameters': {},
                        'protocol': 'tcp',
                        'logical_source_port': src_port['port']['id']
                    }) as fc, self.port_chain(port_chain={
                        'port_pair_groups': [ppg['port_pair_group']['id']],
                        'flow_classifiers': [
                            fc['flow_classifier']['id']
                        ]
                    }) as pc:
                        pc_context = sfc_ctx.PortChainContext(
                            self.sfc_plugin, self.ctx,
                            pc['port_chain']
                        )
                        self.driver.create_port_chain(pc_context)

                        updates = {
                            'flow_classifiers': []
                        }
                        req = self.new_update_request(
                            'port_chains', {'port_chain': updates},
                            pc['port_chain']['id']
                        )
                        res = req.get_response(self.ext_api)
                        pc2 = self.deserialize(
                            self.fmt, res
                        )
                        pc_context = sfc_ctx.PortChainContext(
                            self.sfc_plugin, self.ctx,
                            pc2['port_chain'],
                            original_portchain=pc['port_chain']
                        )

                        self.mocked_push_rpc.reset_mock()
                        self.driver.update_port_chain(pc_context)

                        expected_chain_hops = [dict(
                            project_id=self._tenant_id,
                            portchain_id=pc['port_chain']['id'],
                            ingress_ppg=ppg['port_pair_group']['id'],
                            egress_ppg=None,
                            ingress_network=None,
                            egress_network=self._dest_net['network']['id'],
                            ingress_gw=self._subnet['subnet']['gateway_ip'],
                            egress_gw=sfc_base.DEST_GATEWAY,
                            ingress_ports=[egress_port['port']['id']],
                            egress_ports=[dest_port['port']['id']],
                            rts=[self._dest_rt],
                            readv_from_rts=[],
                            readv_to_rt=None,
                            redirect_rts=[],
                            classifiers=None,
                            reverse_hop=False
                        ), dict(
                            project_id=self._tenant_id,
                            portchain_id=pc['port_chain']['id'],
                            ingress_ppg=None,
                            egress_ppg=ppg['port_pair_group']['id'],
                            ingress_network=self._src_net['network']['id'],
                            egress_network=None,
                            ingress_gw=sfc_base.SRC_GATEWAY,
                            egress_gw=self._subnet['subnet']['gateway_ip'],
                            ingress_ports=[src_port['port']['id']],
                            egress_ports=[ingress_port['port']['id']],
                            rts=['64512:5000'],
                            readv_from_rts=[],
                            readv_to_rt=None,
                            redirect_rts=['64512:5000'],
                            attract_to_rt='64512:5001',
                            classifiers=[
                                dict(protocol='tcp',
                                     sourcePrefix=sfc_base.SRC_CIDR,
                                     sourcePort='100:200',
                                     destinationPrefix=sfc_base.DEST_CIDR,
                                     destinationPort='300:400')],
                            reverse_hop=False
                        )]

                        self.assertEqual(self.mocked_push_rpc.call_count, 1)
                        self._validate_chain_hop_rpc_call(0,
                                                          rpc_events.DELETED,
                                                          expected_chain_hops)

    def test_update_port_chain_replace_flow_classifier(self):
        with self.port(
            name='src_port1',
            device_owner='compute:None',
            device_id='test',
            network_id=self._src_net['network']['id']
        ) as src_port1, self.port(
            name='src_port2',
            device_owner='compute:None',
            device_id='test',
            network_id=self._src_net['network']['id']
        ) as src_port2, self.port(
            name='ingress_port',
            device_owner='compute:None',
            device_id='test'
        ) as ingress_port, self.port(
            name='egress_port',
            device_owner='compute:None',
            device_id='test'
        ) as egress_port:
            with self.port_pair(port_pair={
                'ingress': ingress_port['port']['id'],
                'egress': egress_port['port']['id']
            }) as pp:
                pp_context = sfc_ctx.PortPairContext(
                    self.sfc_plugin, self.ctx,
                    pp['port_pair']
                )
                self.driver.create_port_pair(pp_context)

                with self.port_pair_group(port_pair_group={
                    'port_pairs': [
                        pp['port_pair']['id']
                    ]
                }) as ppg:
                    ppg_context = sfc_ctx.PortPairGroupContext(
                        self.sfc_plugin, self.ctx,
                        ppg['port_pair_group']
                    )
                    self.driver.create_port_pair_group(ppg_context)

                    with self.flow_classifier(flow_classifier={
                        'logical_source_port': src_port1['port']['id'],
                        'destination_ip_prefix': sfc_base.DEST_CIDR
                    }) as fc1, self.flow_classifier(flow_classifier={
                        'logical_source_port': src_port2['port']['id'],
                        'destination_ip_prefix': sfc_base.DEST_CIDR
                    }) as fc2, self.port_chain(port_chain={
                        'port_pair_groups': [ppg['port_pair_group']['id']],
                        'flow_classifiers': [
                            fc1['flow_classifier']['id']
                        ]
                    }) as pc:
                        pc_context = sfc_ctx.PortChainContext(
                            self.sfc_plugin, self.ctx,
                            pc['port_chain']
                        )
                        self.driver.create_port_chain(pc_context)

                        updates = {
                            'flow_classifiers': [
                                fc2['flow_classifier']['id']
                            ]
                        }
                        req = self.new_update_request(
                            'port_chains', {'port_chain': updates},
                            pc['port_chain']['id']
                        )
                        res = req.get_response(self.ext_api)
                        pc2 = self.deserialize(
                            self.fmt, res
                        )
                        pc_context = sfc_ctx.PortChainContext(
                            self.sfc_plugin, self.ctx,
                            pc2['port_chain'],
                            original_portchain=pc['port_chain']
                        )
                        self.assertRaises(sfc_exc.SfcDriverError,
                                          self.driver.update_port_chain,
                                          pc_context)

    def test_update_port_chain_add_port_pair_group(self):
        with self.port(
            name='src_port',
            device_owner='compute:None',
            device_id='test',
            network_id=self._src_net['network']['id']
        ) as src_port, self.port(
            name='ingress_port1',
            device_owner='compute:None',
            device_id='test'
        ) as ingress_port1, self.port(
            name='egress_port1',
            device_owner='compute:None',
            device_id='test'
        ) as egress_port1, self.port(
            name='ingress_port2',
            device_owner='compute:None',
            device_id='test'
        ) as ingress_port2, self.port(
            name='egress_port2',
            device_owner='compute:None',
            device_id='test'
        ) as egress_port2, self.port(
            name='dest_port',
            device_owner='compute:None',
            device_id='test',
            network_id=self._dest_net['network']['id']
        ) as dest_port:
            with self.port_pair(port_pair={
                'ingress': ingress_port1['port']['id'],
                'egress': egress_port1['port']['id']
            }) as pp1, self.port_pair(port_pair={
                'ingress': ingress_port2['port']['id'],
                'egress': egress_port2['port']['id']
            }) as pp2:
                pp1_context = sfc_ctx.PortPairContext(
                    self.sfc_plugin, self.ctx,
                    pp1['port_pair']
                )
                self.driver.create_port_pair(pp1_context)
                pp2_context = sfc_ctx.PortPairContext(
                    self.sfc_plugin, self.ctx,
                    pp2['port_pair']
                )
                self.driver.create_port_pair(pp2_context)

                with self.port_pair_group(port_pair_group={
                    'port_pairs': [
                        pp1['port_pair']['id']
                    ]
                }) as ppg1, self.port_pair_group(port_pair_group={
                    'port_pairs': [
                        pp2['port_pair']['id']
                    ]
                }) as ppg2:
                    ppg1_context = sfc_ctx.PortPairGroupContext(
                        self.sfc_plugin, self.ctx,
                        ppg1['port_pair_group']
                    )
                    self.driver.create_port_pair_group(ppg1_context)
                    ppg2_context = sfc_ctx.PortPairGroupContext(
                        self.sfc_plugin, self.ctx,
                        ppg2['port_pair_group']
                    )
                    self.driver.create_port_pair_group(ppg2_context)

                    with self.flow_classifier(flow_classifier={
                        'source_port_range_min': 100,
                        'source_port_range_max': 200,
                        'destination_port_range_min': 300,
                        'destination_port_range_max': 400,
                        'ethertype': 'IPv4',
                        'source_ip_prefix': sfc_base.SRC_CIDR,
                        'destination_ip_prefix': sfc_base.DEST_CIDR,
                        'l7_parameters': {},
                        'protocol': 'tcp',
                        'logical_source_port': src_port['port']['id']
                    }) as fc, self.port_chain(port_chain={
                        'name': 'test_chain',
                        'port_pair_groups': [
                            ppg1['port_pair_group']['id']
                        ],
                        'flow_classifiers': [
                            fc['flow_classifier']['id']
                        ]
                    }) as pc:
                        pc_context = sfc_ctx.PortChainContext(
                            self.sfc_plugin, self.ctx,
                            pc['port_chain']
                        )
                        self.driver.create_port_chain(pc_context)

                        updates = {
                            'port_pair_groups': [
                                ppg1['port_pair_group']['id'],
                                ppg2['port_pair_group']['id']
                            ]
                        }
                        req = self.new_update_request(
                            'port_chains', {'port_chain': updates},
                            pc['port_chain']['id']
                        )
                        res = req.get_response(self.ext_api)
                        pc2 = self.deserialize(
                            self.fmt, res
                        )
                        pc_context = sfc_ctx.PortChainContext(
                            self.sfc_plugin, self.ctx,
                            pc2['port_chain'],
                            original_portchain=pc['port_chain']
                        )

                        self.mocked_push_rpc.reset_mock()
                        self.driver.update_port_chain(pc_context)

                        expected_deleted_hops = [dict(
                            project_id=self._tenant_id,
                            portchain_id=pc['port_chain']['id'],
                            ingress_ppg=ppg1['port_pair_group']['id'],
                            egress_ppg=None,
                            ingress_network=None,
                            egress_network=self._dest_net['network']['id'],
                            ingress_gw=self._subnet['subnet']['gateway_ip'],
                            egress_gw=sfc_base.DEST_GATEWAY,
                            ingress_ports=[egress_port1['port']['id']],
                            egress_ports=[dest_port['port']['id']],
                            rts=[self._dest_rt],
                            readv_from_rts=[],
                            readv_to_rt=None,
                            redirect_rts=[],
                            classifiers=None,
                            reverse_hop=False
                        ), dict(
                            project_id=self._tenant_id,
                            portchain_id=pc['port_chain']['id'],
                            ingress_ppg=None,
                            egress_ppg=ppg1['port_pair_group']['id'],
                            ingress_network=self._src_net['network']['id'],
                            egress_network=None,
                            ingress_gw=sfc_base.SRC_GATEWAY,
                            egress_gw=self._subnet['subnet']['gateway_ip'],
                            ingress_ports=[src_port['port']['id']],
                            egress_ports=[ingress_port1['port']['id']],
                            rts=['64512:5000'],
                            readv_from_rts=[],
                            readv_to_rt=None,
                            redirect_rts=['64512:5000'],
                            attract_to_rt='64512:5001',
                            classifiers=[
                                dict(protocol='tcp',
                                     sourcePrefix=sfc_base.SRC_CIDR,
                                     sourcePort='100:200',
                                     destinationPrefix=sfc_base.DEST_CIDR,
                                     destinationPort='300:400')],
                            reverse_hop=False
                        )]

                        expected_created_hops = [dict(
                            project_id=self._tenant_id,
                            portchain_id=pc['port_chain']['id'],
                            ingress_ppg=ppg2['port_pair_group']['id'],
                            egress_ppg=None,
                            ingress_network=None,
                            egress_network=self._dest_net['network']['id'],
                            ingress_gw=self._subnet['subnet']['gateway_ip'],
                            egress_gw=sfc_base.DEST_GATEWAY,
                            ingress_ports=[egress_port2['port']['id']],
                            egress_ports=[dest_port['port']['id']],
                            rts=[self._dest_rt],
                            readv_from_rts=[],
                            readv_to_rt=None,
                            redirect_rts=[],
                            classifiers=None,
                            reverse_hop=False
                        ), dict(
                            project_id=self._tenant_id,
                            portchain_id=pc['port_chain']['id'],
                            ingress_ppg=ppg1['port_pair_group']['id'],
                            egress_ppg=ppg2['port_pair_group']['id'],
                            ingress_network=None,
                            egress_network=None,
                            ingress_gw=self._subnet['subnet']['gateway_ip'],
                            egress_gw=self._subnet['subnet']['gateway_ip'],
                            ingress_ports=[egress_port1['port']['id']],
                            egress_ports=[ingress_port2['port']['id']],
                            rts=['64512:5000'],
                            readv_from_rts=[],
                            readv_to_rt=None,
                            redirect_rts=['64512:5000'],
                            attract_to_rt='64512:5001',
                            classifiers=[
                                dict(protocol='tcp',
                                     sourcePrefix=sfc_base.SRC_CIDR,
                                     sourcePort='100:200',
                                     destinationPrefix=sfc_base.DEST_CIDR,
                                     destinationPort='300:400')],
                            reverse_hop=False
                        ), dict(
                            project_id=self._tenant_id,
                            portchain_id=pc['port_chain']['id'],
                            ingress_ppg=None,
                            egress_ppg=ppg1['port_pair_group']['id'],
                            ingress_network=self._src_net['network']['id'],
                            egress_network=None,
                            ingress_gw=sfc_base.SRC_GATEWAY,
                            egress_gw=self._subnet['subnet']['gateway_ip'],
                            ingress_ports=[src_port['port']['id']],
                            egress_ports=[ingress_port1['port']['id']],
                            rts=['64512:5002'],
                            readv_from_rts=[],
                            readv_to_rt=None,
                            redirect_rts=['64512:5002'],
                            attract_to_rt='64512:5003',
                            classifiers=[
                                dict(protocol='tcp',
                                     sourcePrefix=sfc_base.SRC_CIDR,
                                     sourcePort='100:200',
                                     destinationPrefix=sfc_base.DEST_CIDR,
                                     destinationPort='300:400')],
                            reverse_hop=False
                        )]

                        self.assertEqual(self.mocked_push_rpc.call_count, 2)
                        self._validate_chain_hop_rpc_call(
                            0,
                            rpc_events.DELETED,
                            expected_deleted_hops
                        )

                        self._validate_chain_hop_rpc_call(
                            1,
                            rpc_events.CREATED,
                            expected_created_hops
                        )

    def test_update_port_chain_delete_port_pair_group(self):
        with self.port(
            name='src_port',
            device_owner='compute:None',
            device_id='test',
            network_id=self._src_net['network']['id']
        ) as src_port, self.port(
            name='ingress_port1',
            device_owner='compute:None',
            device_id='test'
        ) as ingress_port1, self.port(
            name='egress_port1',
            device_owner='compute:None',
            device_id='test'
        ) as egress_port1, self.port(
            name='ingress_port2',
            device_owner='compute:None',
            device_id='test'
        ) as ingress_port2, self.port(
            name='egress_port2',
            device_owner='compute:None',
            device_id='test'
        ) as egress_port2, self.port(
            name='dest_port',
            device_owner='compute:None',
            device_id='test',
            network_id=self._dest_net['network']['id']
        ) as dest_port:
            with self.port_pair(port_pair={
                'ingress': ingress_port1['port']['id'],
                'egress': egress_port1['port']['id']
            }) as pp1, self.port_pair(port_pair={
                'ingress': ingress_port2['port']['id'],
                'egress': egress_port2['port']['id']
            }) as pp2:
                pp1_context = sfc_ctx.PortPairContext(
                    self.sfc_plugin, self.ctx,
                    pp1['port_pair']
                )
                self.driver.create_port_pair(pp1_context)
                pp2_context = sfc_ctx.PortPairContext(
                    self.sfc_plugin, self.ctx,
                    pp2['port_pair']
                )
                self.driver.create_port_pair(pp2_context)

                with self.port_pair_group(port_pair_group={
                    'port_pairs': [
                        pp1['port_pair']['id']
                    ]
                }) as ppg1, self.port_pair_group(port_pair_group={
                    'port_pairs': [
                        pp2['port_pair']['id']
                    ]
                }) as ppg2:
                    ppg1_context = sfc_ctx.PortPairGroupContext(
                        self.sfc_plugin, self.ctx,
                        ppg1['port_pair_group']
                    )
                    self.driver.create_port_pair_group(ppg1_context)
                    ppg2_context = sfc_ctx.PortPairGroupContext(
                        self.sfc_plugin, self.ctx,
                        ppg2['port_pair_group']
                    )
                    self.driver.create_port_pair_group(ppg2_context)

                    with self.flow_classifier(flow_classifier={
                        'source_port_range_min': 100,
                        'source_port_range_max': 200,
                        'destination_port_range_min': 300,
                        'destination_port_range_max': 400,
                        'ethertype': 'IPv4',
                        'source_ip_prefix': sfc_base.SRC_CIDR,
                        'destination_ip_prefix': sfc_base.DEST_CIDR,
                        'l7_parameters': {},
                        'protocol': 'tcp',
                        'logical_source_port': src_port['port']['id']
                    }) as fc, self.port_chain(port_chain={
                        'name': 'test_chain',
                        'port_pair_groups': [
                            ppg1['port_pair_group']['id'],
                            ppg2['port_pair_group']['id']
                        ],
                        'flow_classifiers': [
                            fc['flow_classifier']['id']
                        ]
                    }) as pc:
                        pc_context = sfc_ctx.PortChainContext(
                            self.sfc_plugin, self.ctx,
                            pc['port_chain']
                        )
                        self.driver.create_port_chain(pc_context)

                        updates = {
                            'port_pair_groups': [
                                ppg1['port_pair_group']['id']
                            ]
                        }
                        req = self.new_update_request(
                            'port_chains', {'port_chain': updates},
                            pc['port_chain']['id']
                        )
                        res = req.get_response(self.ext_api)
                        pc2 = self.deserialize(
                            self.fmt, res
                        )
                        pc_context = sfc_ctx.PortChainContext(
                            self.sfc_plugin, self.ctx,
                            pc2['port_chain'],
                            original_portchain=pc['port_chain']
                        )

                        self.mocked_push_rpc.reset_mock()
                        self.driver.update_port_chain(pc_context)

                        expected_deleted_hops = [dict(
                            project_id=self._tenant_id,
                            portchain_id=pc['port_chain']['id'],
                            ingress_ppg=ppg2['port_pair_group']['id'],
                            egress_ppg=None,
                            ingress_network=None,
                            egress_network=self._dest_net['network']['id'],
                            ingress_gw=self._subnet['subnet']['gateway_ip'],
                            egress_gw=sfc_base.DEST_GATEWAY,
                            ingress_ports=[egress_port2['port']['id']],
                            egress_ports=[dest_port['port']['id']],
                            rts=[self._dest_rt],
                            readv_from_rts=[],
                            readv_to_rt=None,
                            redirect_rts=[],
                            classifiers=None,
                            reverse_hop=False
                        ), dict(
                            project_id=self._tenant_id,
                            portchain_id=pc['port_chain']['id'],
                            ingress_ppg=ppg1['port_pair_group']['id'],
                            egress_ppg=ppg2['port_pair_group']['id'],
                            ingress_network=None,
                            egress_network=None,
                            ingress_gw=self._subnet['subnet']['gateway_ip'],
                            egress_gw=self._subnet['subnet']['gateway_ip'],
                            ingress_ports=[egress_port1['port']['id']],
                            egress_ports=[ingress_port2['port']['id']],
                            rts=['64512:5000'],
                            readv_from_rts=[],
                            readv_to_rt=None,
                            redirect_rts=['64512:5000'],
                            attract_to_rt='64512:5001',
                            classifiers=[
                                dict(protocol='tcp',
                                     sourcePrefix=sfc_base.SRC_CIDR,
                                     sourcePort='100:200',
                                     destinationPrefix=sfc_base.DEST_CIDR,
                                     destinationPort='300:400')],
                            reverse_hop=False
                        ), dict(
                            project_id=self._tenant_id,
                            portchain_id=pc['port_chain']['id'],
                            ingress_ppg=None,
                            egress_ppg=ppg1['port_pair_group']['id'],
                            ingress_network=self._src_net['network']['id'],
                            egress_network=None,
                            ingress_gw=sfc_base.SRC_GATEWAY,
                            egress_gw=self._subnet['subnet']['gateway_ip'],
                            ingress_ports=[src_port['port']['id']],
                            egress_ports=[ingress_port1['port']['id']],
                            rts=['64512:5002'],
                            readv_from_rts=[],
                            readv_to_rt=None,
                            redirect_rts=['64512:5002'],
                            attract_to_rt='64512:5003',
                            classifiers=[
                                dict(protocol='tcp',
                                     sourcePrefix=sfc_base.SRC_CIDR,
                                     sourcePort='100:200',
                                     destinationPrefix=sfc_base.DEST_CIDR,
                                     destinationPort='300:400')],
                            reverse_hop=False
                        )]

                        expected_created_hops = [dict(
                            project_id=self._tenant_id,
                            portchain_id=pc['port_chain']['id'],
                            ingress_ppg=ppg1['port_pair_group']['id'],
                            egress_ppg=None,
                            ingress_network=None,
                            egress_network=self._dest_net['network']['id'],
                            ingress_gw=self._subnet['subnet']['gateway_ip'],
                            egress_gw=sfc_base.DEST_GATEWAY,
                            ingress_ports=[egress_port1['port']['id']],
                            egress_ports=[dest_port['port']['id']],
                            rts=[self._dest_rt],
                            readv_from_rts=[],
                            readv_to_rt=None,
                            redirect_rts=[],
                            classifiers=None,
                            reverse_hop=False
                        ), dict(
                            project_id=self._tenant_id,
                            portchain_id=pc['port_chain']['id'],
                            ingress_ppg=None,
                            egress_ppg=ppg1['port_pair_group']['id'],
                            ingress_network=self._src_net['network']['id'],
                            egress_network=None,
                            ingress_gw=sfc_base.SRC_GATEWAY,
                            egress_gw=self._subnet['subnet']['gateway_ip'],
                            ingress_ports=[src_port['port']['id']],
                            egress_ports=[ingress_port1['port']['id']],
                            rts=['64512:5002'],
                            readv_from_rts=[],
                            readv_to_rt=None,
                            redirect_rts=['64512:5002'],
                            attract_to_rt='64512:5003',
                            classifiers=[
                                dict(protocol='tcp',
                                     sourcePrefix=sfc_base.SRC_CIDR,
                                     sourcePort='100:200',
                                     destinationPrefix=sfc_base.DEST_CIDR,
                                     destinationPort='300:400')],
                            reverse_hop=False
                        )]

                        self.assertEqual(self.mocked_push_rpc.call_count, 2)
                        self._validate_chain_hop_rpc_call(
                            0,
                            rpc_events.DELETED,
                            expected_deleted_hops
                        )

                        self._validate_chain_hop_rpc_call(
                            1,
                            rpc_events.CREATED,
                            expected_created_hops
                        )

    def test_update_port_chain_replace_port_pair_group(self):
        with self.port(
            name='src_port',
            device_owner='compute:None',
            device_id='test',
            network_id=self._src_net['network']['id']
        ) as src_port, self.port(
            name='ingress_port1',
            device_owner='compute:None',
            device_id='test'
        ) as ingress_port1, self.port(
            name='egress_port1',
            device_owner='compute:None',
            device_id='test'
        ) as egress_port1, self.port(
            name='ingress_port2',
            device_owner='compute:None',
            device_id='test'
        ) as ingress_port2, self.port(
            name='egress_port2',
            device_owner='compute:None',
            device_id='test'
        ) as egress_port2, self.port(
            name='dest_port',
            device_owner='compute:None',
            device_id='test',
            network_id=self._dest_net['network']['id']
        ) as dest_port:
            with self.port_pair(port_pair={
                'ingress': ingress_port1['port']['id'],
                'egress': egress_port1['port']['id']
            }) as pp1, self.port_pair(port_pair={
                'ingress': ingress_port2['port']['id'],
                'egress': egress_port2['port']['id']
            }) as pp2:
                pp1_context = sfc_ctx.PortPairContext(
                    self.sfc_plugin, self.ctx,
                    pp1['port_pair']
                )
                self.driver.create_port_pair(pp1_context)
                pp2_context = sfc_ctx.PortPairContext(
                    self.sfc_plugin, self.ctx,
                    pp2['port_pair']
                )
                self.driver.create_port_pair(pp2_context)

                with self.port_pair_group(port_pair_group={
                    'port_pairs': [
                        pp1['port_pair']['id']
                    ]
                }) as ppg1, self.port_pair_group(port_pair_group={
                    'port_pairs': [
                        pp2['port_pair']['id']
                    ]
                }) as ppg2:
                    ppg1_context = sfc_ctx.PortPairGroupContext(
                        self.sfc_plugin, self.ctx,
                        ppg1['port_pair_group']
                    )
                    self.driver.create_port_pair_group(ppg1_context)
                    ppg2_context = sfc_ctx.PortPairGroupContext(
                        self.sfc_plugin, self.ctx,
                        ppg2['port_pair_group']
                    )
                    self.driver.create_port_pair_group(ppg2_context)

                    with self.flow_classifier(flow_classifier={
                        'source_port_range_min': 100,
                        'source_port_range_max': 200,
                        'destination_port_range_min': 300,
                        'destination_port_range_max': 400,
                        'ethertype': 'IPv4',
                        'source_ip_prefix': sfc_base.SRC_CIDR,
                        'destination_ip_prefix': sfc_base.DEST_CIDR,
                        'l7_parameters': {},
                        'protocol': 'tcp',
                        'logical_source_port': src_port['port']['id']
                    }) as fc, self.port_chain(port_chain={
                        'name': 'test_chain',
                        'port_pair_groups': [
                            ppg1['port_pair_group']['id']
                        ],
                        'flow_classifiers': [
                            fc['flow_classifier']['id']
                        ]
                    }) as pc:
                        pc_context = sfc_ctx.PortChainContext(
                            self.sfc_plugin, self.ctx,
                            pc['port_chain']
                        )
                        self.driver.create_port_chain(pc_context)

                        updates = {
                            'port_pair_groups': [
                                ppg2['port_pair_group']['id']
                            ]
                        }
                        req = self.new_update_request(
                            'port_chains', {'port_chain': updates},
                            pc['port_chain']['id']
                        )
                        res = req.get_response(self.ext_api)
                        pc2 = self.deserialize(
                            self.fmt, res
                        )
                        pc_context = sfc_ctx.PortChainContext(
                            self.sfc_plugin, self.ctx,
                            pc2['port_chain'],
                            original_portchain=pc['port_chain']
                        )

                        self.mocked_push_rpc.reset_mock()
                        self.driver.update_port_chain(pc_context)

                        expected_deleted_hops = [dict(
                            project_id=self._tenant_id,
                            portchain_id=pc['port_chain']['id'],
                            ingress_ppg=ppg1['port_pair_group']['id'],
                            egress_ppg=None,
                            ingress_network=None,
                            egress_network=self._dest_net['network']['id'],
                            ingress_gw=self._subnet['subnet']['gateway_ip'],
                            egress_gw=sfc_base.DEST_GATEWAY,
                            ingress_ports=[egress_port1['port']['id']],
                            egress_ports=[dest_port['port']['id']],
                            rts=[self._dest_rt],
                            readv_from_rts=[],
                            readv_to_rt=None,
                            redirect_rts=[],
                            classifiers=None,
                            reverse_hop=False
                        ), dict(
                            project_id=self._tenant_id,
                            portchain_id=pc['port_chain']['id'],
                            ingress_ppg=None,
                            egress_ppg=ppg1['port_pair_group']['id'],
                            ingress_network=self._src_net['network']['id'],
                            egress_network=None,
                            ingress_gw=sfc_base.SRC_GATEWAY,
                            egress_gw=self._subnet['subnet']['gateway_ip'],
                            ingress_ports=[src_port['port']['id']],
                            egress_ports=[ingress_port1['port']['id']],
                            rts=['64512:5000'],
                            readv_from_rts=[],
                            readv_to_rt=None,
                            redirect_rts=['64512:5000'],
                            attract_to_rt='64512:5001',
                            classifiers=[
                                dict(protocol='tcp',
                                     sourcePrefix=sfc_base.SRC_CIDR,
                                     sourcePort='100:200',
                                     destinationPrefix=sfc_base.DEST_CIDR,
                                     destinationPort='300:400')],
                            reverse_hop=False
                        )]

                        expected_created_hops = [dict(
                            project_id=self._tenant_id,
                            portchain_id=pc['port_chain']['id'],
                            ingress_ppg=ppg2['port_pair_group']['id'],
                            egress_ppg=None,
                            ingress_network=None,
                            egress_network=self._dest_net['network']['id'],
                            ingress_gw=self._subnet['subnet']['gateway_ip'],
                            egress_gw=sfc_base.DEST_GATEWAY,
                            ingress_ports=[egress_port2['port']['id']],
                            egress_ports=[dest_port['port']['id']],
                            rts=[self._dest_rt],
                            readv_from_rts=[],
                            readv_to_rt=None,
                            redirect_rts=[],
                            classifiers=None,
                            reverse_hop=False
                        ), dict(
                            project_id=self._tenant_id,
                            portchain_id=pc['port_chain']['id'],
                            ingress_ppg=None,
                            egress_ppg=ppg2['port_pair_group']['id'],
                            ingress_network=self._src_net['network']['id'],
                            egress_network=None,
                            ingress_gw=sfc_base.SRC_GATEWAY,
                            egress_gw=self._subnet['subnet']['gateway_ip'],
                            ingress_ports=[src_port['port']['id']],
                            egress_ports=[ingress_port2['port']['id']],
                            rts=['64512:5002'],
                            readv_from_rts=[],
                            readv_to_rt=None,
                            redirect_rts=['64512:5002'],
                            attract_to_rt='64512:5003',
                            classifiers=[
                                dict(protocol='tcp',
                                     sourcePrefix=sfc_base.SRC_CIDR,
                                     sourcePort='100:200',
                                     destinationPrefix=sfc_base.DEST_CIDR,
                                     destinationPort='300:400')],
                            reverse_hop=False
                        )]

                        self.assertEqual(self.mocked_push_rpc.call_count, 2)
                        self._validate_chain_hop_rpc_call(
                            0,
                            rpc_events.DELETED,
                            expected_deleted_hops
                        )

                        self._validate_chain_hop_rpc_call(
                            1,
                            rpc_events.CREATED,
                            expected_created_hops
                        )

    def test_delete_port_chain(self):
        with self.port_pair_group(
            port_pair_group={'name': 'test_ppg'}
        ) as ppg:
            ppg_context = sfc_ctx.PortPairGroupContext(
                self.sfc_plugin, self.ctx,
                ppg['port_pair_group']
            )
            self.driver.create_port_pair_group(ppg_context)
            with self.port_chain(port_chain={
                'name': 'test_chain',
                'port_pair_groups': [
                    ppg['port_pair_group']['id']
                ]
            }) as pc:
                pc_context = sfc_ctx.PortChainContext(
                    self.sfc_plugin, self.ctx,
                    pc['port_chain']
                )
                self.driver.create_port_chain(pc_context)

                self.mocked_push_rpc.reset_mock()
                self.driver.delete_port_chain(pc_context)

                self.mocked_push_rpc.assert_not_called()

    def test_delete_port_chain_with_port_pairs(self):
        with self.port(
            name='ingress_port',
            device_owner='compute:None',
            device_id='test'
        ) as ingress_port, self.port(
            name='egress_port',
            device_owner='compute:None',
            device_id='test'
        ) as egress_port:
            with self.port_pair(port_pair={
                'ingress': ingress_port['port']['id'],
                'egress': egress_port['port']['id']
            }) as pp:
                pp_context = sfc_ctx.PortPairContext(
                    self.sfc_plugin, self.ctx,
                    pp['port_pair']
                )
                self.driver.create_port_pair(pp_context)
                with self.port_pair_group(port_pair_group={
                    'port_pairs': [pp['port_pair']['id']]
                }) as ppg:
                    ppg_context = sfc_ctx.PortPairGroupContext(
                        self.sfc_plugin, self.ctx,
                        ppg['port_pair_group']
                    )
                    self.driver.create_port_pair_group(ppg_context)
                    with self.port_chain(port_chain={
                        'name': 'test_chain',
                        'port_pair_groups': [
                            ppg['port_pair_group']['id']
                        ]
                    }) as pc:
                        pc_context = sfc_ctx.PortChainContext(
                            self.sfc_plugin, self.ctx,
                            pc['port_chain']
                        )
                        self.driver.create_port_chain(pc_context)

                        self.mocked_push_rpc.reset_mock()
                        self.driver.delete_port_chain(pc_context)

                        self.mocked_push_rpc.assert_not_called()

    def test_delete_port_chain_with_flow_classifiers(self):
        with self.port_pair_group(
            port_pair_group={'name': 'test_ppg'}
        ) as ppg:
            ppg_context = sfc_ctx.PortPairGroupContext(
                self.sfc_plugin, self.ctx,
                ppg['port_pair_group']
            )
            self.driver.create_port_pair_group(ppg_context)
            with self.flow_classifier(flow_classifier={
                'source_port_range_min': 100,
                'source_port_range_max': 200,
                'destination_port_range_min': 300,
                'destination_port_range_max': 400,
                'ethertype': 'IPv4',
                'source_ip_prefix': sfc_base.SRC_CIDR,
                'destination_ip_prefix': sfc_base.DEST_CIDR,
                'l7_parameters': {},
                'protocol': 'tcp'
            }) as fc, self.port_chain(port_chain={
                'name': 'test_chain',
                'port_pair_groups': [
                    ppg['port_pair_group']['id']
                ],
                'flow_classifiers': [
                    fc['flow_classifier']['id']
                ]
            }) as pc:
                pc_context = sfc_ctx.PortChainContext(
                    self.sfc_plugin, self.ctx,
                    pc['port_chain']
                )
                self.driver.create_port_chain(pc_context)

                self.mocked_push_rpc.reset_mock()
                self.driver.delete_port_chain(pc_context)

                self.mocked_push_rpc.assert_not_called()

    def test_delete_port_chain_with_flow_classifiers_port_pairs(self):
        with self.port(
            name='src_port',
            device_owner='compute:None',
            device_id='test',
            network_id=self._src_net['network']['id']
        ) as src_port, self.port(
            name='ingress_port',
            device_owner='compute:None',
            device_id='test'
        ) as ingress_port, self.port(
            name='egress_port',
            device_owner='compute:None',
            device_id='test'
        ) as egress_port, self.port(
            name='dest_port',
            device_owner='compute:None',
            device_id='test',
            network_id=self._dest_net['network']['id']
        ) as dest_port:
            with self.port_pair(port_pair={
                'ingress': ingress_port['port']['id'],
                'egress': egress_port['port']['id']
            }) as pp:
                pp_context = sfc_ctx.PortPairContext(
                    self.sfc_plugin, self.ctx,
                    pp['port_pair']
                )
                self.driver.create_port_pair(pp_context)
                with self.port_pair_group(port_pair_group={
                    'port_pairs': [
                        pp['port_pair']['id']
                    ]
                }) as ppg:
                    ppg_context = sfc_ctx.PortPairGroupContext(
                        self.sfc_plugin, self.ctx,
                        ppg['port_pair_group']
                    )
                    self.driver.create_port_pair_group(ppg_context)
                    with self.flow_classifier(flow_classifier={
                        'source_port_range_min': 100,
                        'source_port_range_max': 200,
                        'destination_port_range_min': 300,
                        'destination_port_range_max': 400,
                        'ethertype': 'IPv4',
                        'source_ip_prefix': sfc_base.SRC_CIDR,
                        'destination_ip_prefix': sfc_base.DEST_CIDR,
                        'l7_parameters': {},
                        'protocol': 'tcp',
                        'logical_source_port': src_port['port']['id']
                    }) as fc, self.port_chain(port_chain={
                        'name': 'test_chain',
                        'port_pair_groups': [
                            ppg['port_pair_group']['id']
                        ],
                        'flow_classifiers': [
                            fc['flow_classifier']['id']
                        ]
                    }) as pc:
                        pc_context = sfc_ctx.PortChainContext(
                            self.sfc_plugin, self.ctx,
                            pc['port_chain']
                        )
                        self.driver.create_port_chain(pc_context)

                        self.mocked_push_rpc.reset_mock()
                        self.driver.delete_port_chain(pc_context)

                        expected_chain_hops = [dict(
                            project_id=self._tenant_id,
                            portchain_id=pc['port_chain']['id'],
                            ingress_ppg=ppg['port_pair_group']['id'],
                            egress_ppg=None,
                            ingress_network=None,
                            egress_network=self._dest_net['network']['id'],
                            ingress_gw=self._subnet['subnet']['gateway_ip'],
                            egress_gw=sfc_base.DEST_GATEWAY,
                            ingress_ports=[egress_port['port']['id']],
                            egress_ports=[dest_port['port']['id']],
                            rts=[self._dest_rt],
                            readv_from_rts=[],
                            readv_to_rt=None,
                            redirect_rts=[],
                            classifiers=None,
                            reverse_hop=False
                        ), dict(
                            project_id=self._tenant_id,
                            portchain_id=pc['port_chain']['id'],
                            ingress_ppg=None,
                            egress_ppg=ppg['port_pair_group']['id'],
                            ingress_network=self._src_net['network']['id'],
                            egress_network=None,
                            ingress_gw=sfc_base.SRC_GATEWAY,
                            egress_gw=self._subnet['subnet']['gateway_ip'],
                            ingress_ports=[src_port['port']['id']],
                            egress_ports=[ingress_port['port']['id']],
                            rts=['64512:5000'],
                            readv_from_rts=[],
                            readv_to_rt=None,
                            redirect_rts=['64512:5000'],
                            attract_to_rt='64512:5001',
                            classifiers=[
                                dict(protocol='tcp',
                                     sourcePrefix=sfc_base.SRC_CIDR,
                                     sourcePort='100:200',
                                     destinationPrefix=sfc_base.DEST_CIDR,
                                     destinationPort='300:400')],
                            reverse_hop=False
                        )]

                        self.assertEqual(self.mocked_push_rpc.call_count, 1)
                        self._validate_chain_hop_rpc_call(0,
                                                          rpc_events.DELETED,
                                                          expected_chain_hops)
